# Copyright (C) 2011-2012 Linaro Limited
#
# Author: Paul Larson <paul.larson@linaro.org>
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.actions.lava
======================

Implementation of generic LAVA test scenario actions
"""

from json_document import bridge
from json_document.document import Document

from lava.core.errors import PreconditionFailed, SemanticProblem
from lava.core.errors import UnrecoverableProblem
from lava.core.history import history
from lava.core.interfaces import IAgentHost, ITestScenarioAction
from lava.core.logging import LoggingMixIn
from lava.core.utils import ObjectWithConfiguration, Decoy

__all__ = ["lava_test_install", "lava_test_run"]


# Singleton for marking things that cannot fail
can_fail = object()


class lava_test_install(LoggingMixIn, ObjectWithConfiguration,
                        ITestScenarioAction):
    """
    Install lava-test and any actual tests that we want to run
    """

    class Configuration(Document):

        document_schema = {
            'type': 'object',
            'properties': {
                'tests': {
                    'type': 'array',
                    'optional': True,
                    'default': [],
                    'unique': True,
                    'items': {
                        'type': 'string'
                    }
                },
                'install_python': {
                    'type': 'array',
                    'optional': True,
                    'default': [],
                    'unique': True,
                    'items': {
                        'type': 'string'
                    }
                },
                'register': {
                    'type': 'array',
                    'optional': True,
                    'default': [],
                    'unique': True,
                    'items': {
                        'type': 'string'
                    }
                },
                'timeout': {
                    'type': 'integer',
                    'optional': True,
                    'default': None
                }
            },
            'additionalProperties': False,
        }

        @bridge.readonly
        def tests(self):
            """
            Array of tests (names) to install
            """

        @bridge.readonly
        def install_python(self):
            """
            Array of pip packages to install
            """

        @bridge.readonly
        def register(self):
            """
            Array of remote tests to register
            """

    def check_precondition(self, state):
        # Ensure the device has finished booting
        if not state.get('boot_done', False):
            raise PreconditionFailed("the device did not boot yet")
        # Ensure the device does not lava-test installed yet
        if state.get('lava_test_installed', False):
            raise PreconditionFailed("lava-test has been already installed")
        # Ensure we are not installing any duplicates
        # XXX: eh, is this even useful?, we cannot call lava_test_install twice
        # anyway
        duplicate_tests = (
            frozenset(self.configuration.tests)
            & frozenset(state.get("installed_tests", [])))
        if duplicate_tests:
            raise PreconditionFailed(
                "duplicate tests installed: {0}".format(
                    ', '.join(duplicate_tests)))

    def emulate(self, controller, state):
        # Ensure we can use the agent API
        if not isinstance(controller, IAgentHost):
            raise SemanticProblem(
                "this action requires controller implementing IAgentHost")
        # Pretend we've installed all the tests
        state['installed_tests'] = self.configuration.tests
        state['lava_test_installed'] = True

    def execute(self, controller, session):
        cmds = self._get_shell_commands(controller, session)
        _run_commands(cmds, controller, session)

    def _get_shell_commands(self, controller, session):
        with self.history.info("Preparing to install tests..."):
            # Update apt cache
            yield "apt-get update", can_fail
            # Install a bunch of packages that we will need
            yield ("apt-get install --yes --force-yes "
                   "bzr "
                   "lsb-release "
                   "python-apt "
                   "python-keyring "
                   "python-pip "
                   "python-setuptools "
                   "python-simplejson "
                   "usbutils ")
        # Install lava-test from pip
        with self.history.info(
            "Installing lava-test on {device}...",
            device=controller.device_desc.device_id):
            yield "pip install lava-test"  # TODO: Add option to customize this
            session.state['lava_test_installed'] = True
            # I don't suppose we need this but what the hell
            yield "lava-test reset"
        # First install any pip packages
        with self.history.debug("Installing additional packages..."):
            for pip_name in self.configuration.install_python:
                with self.history.info("Installing python package: {pip_name}",
                                       pip_name=pip_name):
                    yield "pip install {0}".format(pip_name)
        # Then register any remote tests
        with self.history.debug("Registering external tests..."):
            for location in self.configuration.register:
                with self.history.info(("Registering external test from "
                                        "{location}"), location=location):
                    yield "lava-test register {0}".format(location)
        # Finally install each test
        with self.history.debug("Installing LAVA tests..."):
            session.state['installed_tests'] = []
            for test_name in self.configuration.tests:
                with self.history.info("Installing test {test_name}",
                                       test_name=test_name):
                    yield "lava-test install {0}".format(test_name)
                    # And remember this in the state
                    session.state['installed_tests'].append(test_name)


class lava_test_run(LoggingMixIn, ObjectWithConfiguration,
                    ITestScenarioAction):
    """
    Run lava-test on the device
    """

    class Configuration(Document):

        document_schema = {
            'type': 'object',
            'properties': {
                'test_name': {
                    'type': 'string'
                },
                'test_options': {
                    'type': 'string',
                    'optional': True,
                    'default': ""
                },
                # TODO: --skip-softwre-context
                # TODO: --skip-hardware-context
                # TODO: --trusted-time
                # TODO: and all the other lava-test options
                'timeout': {
                    'type': 'integer',
                    'optional': True,
                    'default': None
                },
            },
            'additionalProperties': False,
        }

        @bridge.readonly
        def test_name(self):
            """
            Name of the lava-test test to run
            """

        @bridge.readonly
        def test_options(self):
            """
            Options to the lava-test test
            """

    def check_precondition(self, state):
        # Ensure the device has finished booting
        if not state.get('boot_done', False):
            raise PreconditionFailed("the device did not boot yet")
        # Ensure the device has lava-test installed
        if not state.get('lava_test_installed', False):
            raise PreconditionFailed("lava-test has not been installed yet")
        # Ensure the test we want to run has been installed
        if self.configuration.test_name not in (
            state.get("installed_tests", [])):
            raise PreconditionFailed(
                "test {0} has not been installed yet".format(
                    self.configuration.test_name))

    def emulate(self, controller, state):
        # Ensure we can use the agent API
        if not isinstance(controller, IAgentHost):
            raise SemanticProblem(
                "this action requires controller implementing IAgentHost")
        # Pretend we ran all the tests
        if 'bundles' not in state:
            state['bundles'] = []
        state['bundles'].append(
            Decoy("bundle from {0}".format(self.configuration.test_name)))

    def execute(self, controller, session):
        if 'bundles' not in session.state:
            session.state['bundles'] = []
        with self.history.info("Running test {test} {options}...",
                               test=self.configuration.test_name,
                               options=self.configuration.test_options):
            cmds = self._get_shell_commands(controller, session)
            _run_commands(cmds, controller, session)
        # XXX: maybe use contextlib.nested?
        with self.history.info("Saving dashboard bundle...") as activity:
            with session.scratch.unique(prefix='bundle.',
                                        suffix='.json') as stream:
                activity.attach(stream)
                stream.writelines(
                    controller.agent.read("/tmp/bundle.json"))
                session.state['bundles'].append(stream.name)

    def _get_shell_commands(self, controller, session):
        # TODO: number each file separately, based on session.state['bundles']
        yield ("lava-test run -o {bundle} {configuration.test_name} "
               "-- {configuration.test_options}").format(
                   bundle="/tmp/bundle.json", configuration=self.configuration)


def _run_commands(cmd_gen, controller, session, all_cannot_fail=True):
    """
    Run a sequence of commands via the agent.
    """
    agent = controller.agent
    for command in cmd_gen:
        if isinstance(command, tuple):
            command, dummy = command
            can_fail = True
        else:
            can_fail = not all_cannot_fail
        with history.info("Invoking command: {command}",
                          command=command) as activity:
            agent.invoke(command)
            code = agent.return_code()
            history.debug("Command finished with error code {code}",
                          code=code)
            # Save stdout
            with session.scratch.open("cmd.%s.out" % agent.sequence) as stream:
                stream.writelines(agent.stdout())
            activity.attach(stream)
            # Save stderr
            with session.scratch.open("cmd.%s.err" % agent.sequence) as stream:
                stream.writelines(agent.stderr())
            activity.attach(stream)
            if code != 0 and not can_fail:
                raise UnrecoverableProblem(
                    "Command %r failed with code %d" % (command, code))
