# Copyright (C) 2011-2012 Linaro Limited
#
# Author: Paul Larson <paul.larson@linaro.org>
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.actions.linaro
========================

Implementation of Linaro-specific test scenario actions
"""

from json_document import bridge
from json_document.document import Document


from lava.core.errors import SemanticProblem, PreconditionFailed
from lava.core.errors import UnrecoverableProblem
from lava.core.interfaces import IAgentHost, IBootable, ITestScenarioAction
from lava.core.logging import LoggingMixIn
from lava.core.utils import ObjectWithConfiguration


__all__ = ["boot_linaro_image"]


class boot_linaro_image(LoggingMixIn, ObjectWithConfiguration,
                        ITestScenarioAction):
    """
    Boot a classical-linux Linaro image.
    """

    class Configuration(Document):

        document_schema = {
            'type': 'object',
            'properties': {
                'shell_prompt': {
                    'type': 'string',
                    'optional': True,
                    'default': 'root@.*:.*#',
                },
                'timeout': {
                    'type': 'integer',
                    'optional': True,
                    'default': 300
                }
            },
            'additionalProperties': False
        }

        @bridge.readonly
        def shell_prompt(self):
            """
            Pattern that describes the shell prompt.
            """

        @bridge.readonly
        def timeout(self):
            """
            Timeout, in seconds, for the boot to finish
            """

    def check_precondition(self, state):
        if 'image' not in state:
            raise PreconditionFailed("there is no image to boot")
        if 'boot_done' in state:
            raise PreconditionFailed("boot is already done, booting twice?")

    def emulate(self, controller, state):
        if not isinstance(controller, IBootable):
            raise SemanticProblem(
                "this action requires controller implementing IBootable")
        if not isinstance(controller, IAgentHost):
            raise SemanticProblem(
                "this action requires controller implementing IAgentHost")
        # Pretend we have booted and started the agent
        state['boot_done'] = True
        state['agent_ready'] = True

    def execute(self, controller, session):
        with self.history.info("Booting device: {device}",
                               device=controller.device_desc.device_id):
            image = session.state['image']
            try:
                controller.boot(
                    session, image, self.configuration.shell_prompt,
                    self.configuration.timeout)
            except Exception as exc:
                raise UnrecoverableProblem(
                    "The device has failed to boot: {0}".format(exc))
            else:
                session.state['boot_done'] = True
        with self.history.info("Starting agent..."):
            try:
                controller.start_agent()
            except Exception as exc:
                raise UnrecoverableProblem(
                    "The agent has failed to start: {0}".format(exc))
            else:
                session.state['agent_ready'] = True
