# Copyright (C) 2011-2012 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.commands.device
=========================
"""

from lava.core.commands import (WorkspaceMixIn, DisplayMixIn, Command,
                                CommandGroup)
from lava.core.errors import CommandError
from lava.core.plugins import PluginLookupError


class device(CommandGroup):
    """
    Interact with test devices
    """

    namespace = 'lava.device.commands'


class list_devices(Command, WorkspaceMixIn, DisplayMixIn):
    """
    display a list of test devices

    Displays a simple list of devices that have been added to the
    current workspace.

    The format of each line is
        {device-id}: {device-name} ({device-backend})
    """

    @classmethod
    def get_name(self):
        return "list"

    def invoke(self):
        self._list_objects(
            self.workspace.environment.devices,
            "{obj.device_id}: {obj.name} ({obj.controller})",
            empty=("You have not defined any devices, "
                   "try 'lava device add --help'"))


class add_device(Command, WorkspaceMixIn, DisplayMixIn):
    """
    add a new test device

    LAVA maintains a registry of 'devices' and their configuration. This
    command adds a new device to that registry. Each device is identified by an
    unique name (you can think of it as a hostname although actual device
    hostname (at runtime) is usually different.

    Devices are managed by device controllers, pieces of code in LAVA that know
    how to control a piece of hardware (such as a development board) or
    software (such as a virtual machine).

    LAVA comes with a few controllers built in. Additional, third party
    extensions, can define more. Interested developers should see the
    lava.core.interfaces.IDeviceController class. To list the controllers that
    you currently have run 'lava device controllers'.Typically you will use one
    of the built-in controllers such as 'qemu' for working with simple virtual
    machines or 'master' for working with common development boards.

    Each controller also requires controller-specific configuration. Some of
    that configuration may be required to operate the device and the 'lava
    device add' command will fail unless you provide it. Other conifiguration
    options may tweak specific aspects. Refer to controller documentation for
    details.

    Examples:

        # Add a virtual Beagle XM:
        $ lava device add beagle01 -c qemu \\
                -C qemu_system=qemu-system-arm \\
                -C qemu_machine=beaglexm
    """

    @classmethod
    def get_name(self):
        return "add"

    @classmethod
    def register_arguments(cls, parser):
        parser.add_argument(
            "device_id",
            metavar="DEVICE-ID",
            help="unique device identifier")
        group = parser.add_argument_group(title="device configuration")
        group.add_argument(
            "-n", "--name", default="",
            help="human readable name, can be changed later")
        group.add_argument(
            "-c", "--controller",
            required=True,
            help="device controller to use (see below)")
        group.add_argument(
            "-C", "--configuration",
            action="append",
            metavar="KEY=VALUE",
            default=[],
            help="arbitrary key-value configuration pair")

    def invoke(self):
        env = self.workspace.environment
        if self.args.device_id in env.devices:
            raise CommandError(
                "Device {0} already exists".format(self.args.device_id))
        env.devices[self.args.device_id] = {}
        device = env.devices[self.args.device_id]
        device.name = self.args.name
        device.controller = self.args.controller
        conf = device.configuration
        for item in self.args.configuration:
            key, value = item.split("=", 1)
            conf[key] = value
        try:
            device.load()
        except PluginLookupError:
            raise CommandError(
                "Unable to find controller: {0}".format(
                    self.args.controller))
        except ValueError as exc:
            raise CommandError("Unable to add device: {0}".format(exc))
        else:
            self.say("Device added")
            self.workspace.save_environment()
