# Copyright (C) 2011-2012 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.commands.scenario
===========================
"""

import getpass
import os
import shutil

from json_document.shortcuts import persistence

from lava.core.commands import WorkspaceMixIn, Command, CommandGroup
from lava.core.errors import SemanticProblem, PreconditionFailed, CommandError
from lava.core.formats import TestScenario
from lava.core.interfaces import ITestScenarioInterpreter


class scenario(CommandGroup):
    """
    Interact with test scenarios
    """

    namespace = "lava.scenario.commands"


class _CommandWithScenario(Command, WorkspaceMixIn):

    @classmethod
    def register_arguments(cls, parser):
        group = parser.add_argument_group(title="test configuration")
        group.add_argument(
            "scenario",
            metavar="SCENARIO",
            help="JSON file with the test scenario")
        group.add_argument(
            "-t", "--target",
            metavar="DEVICE-ID",
            default=None,
            required=True,
            help="Run the scenario on the specified DEVICE-ID")

    def get_scenario(self):
        """
        Get the TestScenario instance that was specified in 'scenario'
        positional argument.
        """
        scenario = TestScenario({})
        try:
            self.logger.info(
                "Loading test scenario from %r", self.args.scenario)
            persistence(
                scenario, self.args.scenario, ignore_missing=False
            ).load()
        except Exception as exc:
            raise CommandError(
                "Scenario {0!r} failed to load: {1}".format(
                    self.args.scenario, exc))
        return scenario

    def get_device(self):
        """
        Get a IDevice instance that was references in --target
        """
        # TODO: add support for reusing target specified in the scenario
        try:
            device_desc = self.workspace.environment.devices[self.args.target]
        except KeyError:
            raise CommandError(
                "Device {0!r} does not exist".format(self.args.target))
        try:
            device = device_desc.load()
        except ValueError as exc:
            raise CommandError(
                "Device {0!r} failed to load: {1}".format(
                    self.args.target, exc))
        # If the device cannot run test scenarios, abort
        if not isinstance(device, ITestScenarioInterpreter):
            raise CommandError(
                "Device {0!r} cannot run test scenarios".format(
                    self.args.target))
        # Otherwise check if we have all the software needed
        try:
            device.check_software(self.workspace.environment)
        except PreconditionFailed as exc:
            raise CommandError(
                "Device {0!r} requires additional software: {1}".format(
                    self.args.target, exc))
        return device

    def emulate_scenario(self, scenario, device):
        try:
            device.emulate_scenario(scenario)
        except (PreconditionFailed, SemanticProblem) as exc:
            raise CommandError(
                "Scenario {0!r} cannot run on device {1!r}: {2}".format(
                    self.args.scenario, self.args.target, exc))

    def prompt_for_credentials(self):
        """
        Prompt the user for sudo password if needed
        """
        credentials = self.workspace.environment['credentials']
        # TODO: promote to environment feature, we should integrate with
        if credentials['sudo-password'].value is None:
            self.say("sudo password is not specified in the environment file")
            password = getpass.getpass("sudo password: ")
            credentials['sudo-password'] = password


class check(_CommandWithScenario):
    """
    Check a test scenario
    """

    def invoke(self):
        scenario = self.get_scenario()
        device = self.get_device()
        self.emulate_scenario(scenario, device)
        self.say("The scenario seems correct")
        self.say("NOTE: LAVA did not check if any of the URLs in your "
                 "scenario actually work. Anything that happens at "
                 "runtime, such as downloading images, installing tests "
                 "and other software, can always fail.")


class run(_CommandWithScenario):
    """
    Run a test scenario
    """

    @classmethod
    def register_arguments(cls, parser):
        super(run, cls).register_arguments(parser)
        group = parser.add_argument_group(title="test result processing")
        group.add_argument(
            "-o", "--output-dir",
            metavar="DIRECTORY",
            required=True,
            default=None,
            help="Save bundles to DIRECTORY")
        # TODO: add support for aggregated bundles where all the test runs are
        # coalesced in one bundle
        group.add_argument_group(title="miscellaneous options")
        group.add_argument(
            "--job-id",
            action='store',
            default=None,
            help=("Set the scheduler job identifier. "
                  "This alters process name for easier debugging"))

    def alter_process_title(self, set_fn, get_fn):
        if self.args.job_id:
            set_fn("{0} [job: {1}]".format(get_fn(), self.args.job_id))

    def invoke(self):
        scenario = self.get_scenario()
        device = self.get_device()
        self.emulate_scenario(scenario, device)
        # Check if we can save results later
        if not os.path.isdir(self.args.output_dir):
            raise CommandError(
                "{0} is not a directory".format(self.args.output_dir))
        # Ensure we have all the secrets
        self.prompt_for_credentials()
        # Run the scenario
        with self.workspace.session() as session:
            session.state['bundles'] = []
            try:
                device.execute_scenario(scenario, session)
                for bundle in session.state['bundles']:
                    shutil.copyfile(bundle,
                                    os.path.join(self.args.output_dir,
                                                 os.path.basename(bundle)))
            finally:
                # Ensure the device is always stopped
                device.stop()
                # XXX: workaround the fact that session is going away and
                # we really want to have attachments there
                if self.args.history:
                    self.args.dispatcher.save_history_now(session.scratch)
