# Copyright (C) 2011-2012 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.commands.software
===========================
"""
import os
import fnmatch

from lava.core.commands import (WorkspaceMixIn, DisplayMixIn, Command,
                                CommandGroup)
from lava.core.errors import CommandError


class software(CommandGroup):
    """
    Configure third party software

    Sometimes LAVA needs to invoke certain external programs. Normally that
    does not require any additional setup. As log as the program is installed
    and can be found on PATH everything should just work out of the box.

    There are certain situations where the user may want to override this
    without changing any other aspects of the system. For example, in a special
    workspace, one could instruct LAVA to use linaro-media-create form some
    custom branch or virtualenv. Another example may be the need to use
    licensed proprietary program that needs certain environment variables to
    function. Even if the program is visible on PATH (which may be tedious as
    such software is rarely properly packaged) it requires the administrator to
    remember about this particular setting. Worse, multiple users interacting
    with one environment can have their own environment variables inherited
    from their SHELL (including overridden PATH) and that might make LAVA
    mysteriously fail for one user while working for everyone else.

    To solve both problems outlined above in a consistent manner, LAVA offers
    explicit configuration option for all programs it has to use to perform
    requested tasks. The configuration can define alternate binary (in that
    case a program 'foo' will not be located using PATH but will always be
    executed with the configured pathname, as stored in the workspace
    environment. In addition, the configuration can add custom environment
    variables that will be set before the program is started.
    """

    namespace = 'lava.software.commands'


class list_software(Command, WorkspaceMixIn, DisplayMixIn):
    """
    display a list of configured external programs
    """

    @classmethod
    def get_name(self):
        return "list"

    def invoke(self):
        if not self.workspace.environment.software:
            self.say("You have not defined any non-standard software "
                     "configuration, try 'lava software use --help'")
        for software in self.workspace.environment.software:
            if software.executable is None:
                self.say("{software}: {exe}{env}",
                         software=software.software_id,
                         exe=(software.executable
                              if software.executable is not None
                              else "looked up on path"),
                         env="*" if software.environment else "")
            else:
                self.say("{obj.software_id}: {obj.executable}", obj=software)


class list_software(Command, WorkspaceMixIn, DisplayMixIn):
    """
    display a list of configured external programs

    The list can be filtered with glob(3) style patterns.
    By default all _configured_ programs are displayed.

    Note that programs without explicit configuration can still exist, and be
    usable for LAVA but they will not be listed here. To configure a specific
    program see 'lava software use --help'

    For example:

        # Show non-standard configuration of programs starting with 'linaro-'
        $ lava software list linaro-*

        # Show all configured programs
        $ lava software list
    """

    @classmethod
    def get_name(self):
        return "list"

    @classmethod
    def register_arguments(cls, parser):
        parser.add_argument(
            "patterns",
            metavar="PROGRAM",
            nargs="*",
            default="*",
            help="glob style program/executable filter")

    def invoke(self):
        if not self.workspace.environment.software:
            self.say("You have not defined any non-standard software "
                     "configuration, try 'lava software use --help'")
        for software in self.workspace.environment.software:
            if not any((fnmatch.fnmatch(software.software_id, pattern)
                        for pattern in self.args.patterns)):
                continue
            if software.executable is None:
                self.say("{obj.software_id}: looked up on path", obj=software)
            else:
                self.say("{obj.software_id}: {obj.executable}", obj=software)
            if software.environment:
                for variable in software.environment:
                    self.say("\t{key}={value}", key=variable.item,
                             value=variable.value)


class use_software(Command, WorkspaceMixIn, DisplayMixIn):
    """
    use non-default configuration for certain program

    You can ask LAVA to always ignore PATH and use explicit executable by
    providing full pathname of the executable.

    For example:

        # Use custom linaro-media-create
        $ lava software use /opt/linaro/linaro-image-tools/bin/linaro-media-create

    You can also ask LAVA to always set additional environment variables before
    starting that program. You don't need to use the full pathname of the
    executable (that would also imply you want to ignore PATH). You can specify
    as many environment variables as you like. If you specify empty environment
    variable it is removed from the 'override'.

    For example:

        # Ask linaro-media-create in /ramdisk to keep temporary data in a ramdisk
        $ lava software use linaro-media-create -E TMPDIR=/ramdisk

        # Undo the action above
        $ lava software use linaro-media-create -E TMPDIR=

    Note that existing configuration for that program will be overridden. Only
    the parts you specify (absolute path or environment variables) are actually
    changed. If in doubt, use `lava software show` or `lava software revert`
    """

    @classmethod
    def get_name(self):
        return "use"

    @classmethod
    def register_arguments(cls, parser):
        parser.add_argument(
            "executable",
            help="executable to use")
        group = parser.add_argument_group(title="program configuration")
        group.add_argument(
            "-E", "--environment",
            action="append",
            metavar="KEY=VALUE",
            default=[],
            help="adds custom environment variable")

    def invoke(self):
        # Access (and load) the environment
        env = self.workspace.environment
        if os.path.isabs(self.args.executable):
            # Treat absolute pathnames as requests to use that executable
            # directly, overriding PATH
            software_id = os.path.basename(self.args.executable)
            executable = self.args.executable
        else:
            # Treat non-absolute pathnames as requests to use the system PATH
            # to find the executable at runtime
            software_id = self.args.executable
            executable = None
        # Access software configuration for this software_id (really program
        # basename)
        soft = env.software[software_id]
        # Save executable location
        if executable is None:
            self.history.info("Will use system PATH to find {software}",
                              software=software_id)
        else:
            self.history.info("Will use {executable} for {software}",
                              executable=executable, software=software_id)
        soft.executable = executable
        # Access software configuration environment variables
        soft_env = soft.environment
        for item in self.args.environment:
            # Treat -E JUST_KEY as -E JUST_KEY=
            try:
                key, value = item.split("=", 1)
            except ValueError:
                key = item
                value = None
            if value:
                # If there was a =VALUE, set it
                self.history.info(
                    "Setting environment variable {key}={value}",
                    key=key, value=value)
                soft_env[key] = value
            else:
                # Otherwise, remove it
                self.history.info(
                    "Removing environment variable {key}", key=key)
                try:
                    del soft_env[key]
                except KeyError:
                    pass
        # Try to "load" the software, this will check for missing binary
        try:
            self.history.info("Checking...")
            soft.load()
        except ValueError as exc:
            raise CommandError(
                "The program does not seem to function: {0}".format(exc))
        else:
            # Save the environment (only done if needed by json-document)
            self.workspace.save_environment()


class revert_software(Command, WorkspaceMixIn, DisplayMixIn):
    """
    remove configuration associated with given software
    This command simply wipes all configuration of the specified program. You
    can use either full absolute pathname or just the name of the executable.

    Example:

        # Return to stock linaro-media-create
        $ lava software revert linaro-media-create
    """

    @classmethod
    def get_name(self):
        return "revert"

    @classmethod
    def register_arguments(cls, parser):
        parser.add_argument(
            "executable",
            metavar="PROGRAM",
            help="executable to revert settings for")

    def invoke(self):
        env = self.workspace.environment
        # Allow users to specify full path or just the executable name
        software_id = os.path.basename(self.args.executable)
        # Abort if there is no configuration for that executable
        if software_id not in env.software:
            raise CommandError("There are no special settings to revert")
        self.history.info("Reverting settings for {software} to default",
                          software=software_id)
        # Revert settings for the specified program. Default settings are empty
        # (for now), so this is equivalent to `del env.software[software_id]`
        env.software[software_id].revert_to_default()
        # Save everything back
        self.workspace.save_environment()
