# Copyright (C) 2011-2012 Linaro Limited
#
# Author: Michael Hudson-Doyle <michael.hudson@linaro.org>
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.controllers.software
==============================

Utilities for handling third party software
"""

import subprocess
import sys

import pexpect

from lava.core.interfaces import ISoftwareController
from lava.core.environment import SoftwareConfiguration
from lava.core.utils import ObjectWithConfiguration
from lava.core.logging import LoggingMixIn


class GenericSoftwareController(ObjectWithConfiguration, LoggingMixIn,
                                ISoftwareController):
    """
    Generic software controller for typical programs
    """

    Configuration = SoftwareConfiguration

    def __init__(self, configuration, executable, sudo_password):
        ObjectWithConfiguration.__init__(self, configuration)
        self._executable = executable
        self._sudo_password = sudo_password
        try:
            self._version = self._probe_version()
            self.history.debug(
                "Program {executable} has version {version!r}",
                executable=self._executable, version=self._version)
        except subprocess.CalledProcessError:
            raise ValueError("incorrect binary")

    @property
    def executable(self):
        """
        Pathname of the executable
        """
        return self.configuration.executable or self._executable

    @property
    def version(self):
        """
        Version of the executable
        """
        return self._version

    def start(self, args, logfile=None, timeout=30, maxread=2000):
        """
        Start the program and return pexpect object
        """
        # env = dict(os.environ)
        env = {"LANG": "C.UTF-8"}
        env.update(self.configuration.environment.value)
        with self.history.debug(("Starting {executable} with arguments: {args}"
                                 " and environment {env}"),
                                executable=self.executable,
                                args=args, env=env) as activity:
            if logfile is None:
                logfile = sys.stdout
            else:
                activity.attach(logfile)
            return pexpect.spawn(self.executable, args, timeout, maxread,
                                 env=env, logfile=logfile)

    def _probe_version(self):
        with self.history.debug("Probing {executable} version...",
                                executable=self.executable):
            return subprocess.check_output(
                [self.executable, '--version'],
                stderr=subprocess.STDOUT).strip()
