# Copyright (C) 2011-2012 Linaro Limited
#
# Author: Michael Hudson-Doyle <michael.hudson@linaro.org>
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.controllers.software.lmc
==================================
"""

import subprocess

import pexpect

from lava.core.controllers.software import GenericSoftwareController


class LinaroMediaCreateController(GenericSoftwareController):
    """
    Wrapper around linaro-media-create.

    This wrapper has helpful high-level methods like generate_image(). It
    also handles obscure details required for automatic image generation
    (when faced with interactive license prompts)
    """

    def generate_image(self, args, **kwargs):
        """
        Run linaro-media-create and accept licenses thrown up in the process.
        """
        # This code is a bit out of control.  It describes a state machine.
        # Each state has a name, a mapping patterns to wait for -> state to
        # move to, a timeout for how long to wait for said pattern and
        # optionally some input to send to l-m-c when you enter the step.

        # The basic outline is this:

        # We wait for l-m-c to actually start.  This has an enormous timeout,
        # because 'cmd' starts with 'flock /var/lock/lava-lmc.lck' and when
        # lots of jobs start at the same time, it can be a long time before the
        # lock is acquired.

        # Once its going, we watch for a couple of key phrases that suggets a
        # license popup has appeared.  The next few states navigate through the
        # dialogs and then accept the license.  The 'say-yes' state has extra
        # fun stuff to try to move to a state where the "<Ok>" button is
        # highlighted before pressing space (the acceptance dialogs are not
        # consistent about whether <Ok> is the default or not!).

        states = {
            'waiting': {
                'expectations': {
                    "linaro-hwpack-install": 'default',
                    '\[sudo\] password for': 'write-password',
                },
                'timeout': 86400,
            },
            'write-password': {
                'input': self._sudo_password + "\n",
                'expectations': {
                    "linaro-hwpack-install": 'default',
                },
                'secret': True,  # this command talks passwords
                'timeout': 30,
            },
            'default': {
                'expectations': {
                    "TI TSPA Software License Agreement": 'accept-tspa',
                    "SNOWBALL CLICK-WRAP": 'accept-snowball',
                    "LIMITED LICENSE AGREEMENT FOR APPLICATION  DEVELOPERS":
                    'accept-snowball',
                },
                'timeout': 3600,
            },
            'accept-tspa': {
                'expectations': {"<Ok>": 'accept-tspa-1'},
                'timeout': 1,
            },
            'accept-tspa-1': {
                'input': "\t ",
                'expectations': {
                    "Accept TI TSPA Software License Agreement": 'say-yes',
                },
                'timeout': 1,
            },
            'say-yes': {
                'expectations': {
                    "  <(Yes|Ok)>": 'say-yes-tab',
                    "\\033\[41m<(Yes|Ok)>": 'say-yes-space',
                },
                'timeout': 1,
            },
            'say-yes-tab': {
                'input': "\t",
                'expectations': {
                    ".": 'say-yes',
                },
                'timeout': 1,
            },
            'say-yes-space': {
                'input': " ",
                'expectations': {
                    ".": 'default',
                },
                'timeout': 1,
            },
            'accept-snowball': {
                'expectations': {"<Ok>": 'accept-snowball-1'},
                'timeout': 1,
            },
            'accept-snowball-1': {
                'input': "\t ",
                'expectations': {
                    "Do you accept": 'say-yes',
                },
                'timeout': 1,
            },
        }
        proc = self.start(args, **kwargs)
        try:
            state = 'waiting'
            while state is not None:
                state_data = states[state]
                patterns = []
                next_state_names = []
                self.logger.debug("Inside l-m-c chat loop (state: %s)", state)
                if 'input' in state_data:
                    if 'secret' in state_data:
                        proc.waitnoecho()
                        proc.setecho(False)
                        self.logger.debug("Sending (secret) to l-m-c")
                    else:
                        self.logger.debug("Sending %r to l-m-c",
                                          state_data['input'])
                    proc.send(state_data['input'])
                    if 'secret' in state_data:
                        proc.setecho(True)
                for pattern, next_state in state_data['expectations'].items():
                    patterns.append(pattern)
                    next_state_names.append(next_state)
                patterns.append(pexpect.EOF)
                next_state_names.append(None)
                self.logger.debug(
                    'Waiting for l-m-c patterns: %r (up to %s seconds)',
                    patterns, state_data['timeout'])
                match_id = proc.expect(patterns, timeout=state_data['timeout'])
                state = next_state_names[match_id]
                self.logger.debug(
                    "Transitioning to l-m-c loop state: %s", state)
        finally:
            proc.close()
        if hasattr(proc, "exitstatus") and proc.exitstatus:
            raise subprocess.CalledProcessError(
                proc.exitstatus, [self.executable] + args)
        if hasattr(proc, "signalstatus") and proc.signalstatus:
            raise subprocess.CalledProcessError(
                -proc.signalstatus, [self.executable] + args)
