# Copyright (C) 2011-2012 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.environment
=====================

Definition of global configuration system for LAVA
(global as in used everywhere, not really global/only-one-per-system)
"""

__all__ = ['Environment']

from json_document import bridge
from json_document.document import Document

from lava.core.interfaces import IDeviceController, ISoftwareController
from lava.core.plugins import PluginLoader, PluginLookupError


class DeviceAttributes(Document):
    """
    Fragment of the Environment document that describes backend-independend
    attributes of this device. Attribtues that appear in this collection are
    regulated by the LAVA development team.  Additional attributes may be
    introduced when the document format changes.
    """

    document_schema = {
        "type": "object",
        "properties": {
            "type": {
                "type": "string"
            }
        },
        "additionalProperties": False
    }

    @bridge.readwrite
    def type(self):
        """
        Name of device type.

        The name is free-form, it only matters when matching test devices from
        the whole environment to a single test scenario. Each organization
        using lava can use custom device names and there is no common naming
        authority. It is advised to use reverse domain name schem to build a
        namespace. For instance Linaro may use a device type such as
        `org.linaro.lava:beagle`
        """


class DeviceDescription(Document):
    """
    Fragment of the Environment documnent that describes a single device.
    """

    _loader = PluginLoader("lava.controllers.devices", IDeviceController)

    document_schema = {
        "type": "object",
        "properties": {
            "name": {
                "type": "string"
            },
            "controller": {
                "type": "string"
            },
            "configuration": {
                "type": "object",
                "optional": True,
                "default": {},
            },
            "attributes": DeviceAttributes,
        }
    }

    @bridge.readwrite
    def name(self):
        """
        Human readable name of this device
        """

    @bridge.readwrite
    def controller(self):
        """
        Name of the controller that should be used to operate this device
        """

    @bridge.fragment
    def configuration(self):
        """
        Device configuration, specific to the backend
        """

    @bridge.fragment
    def attributes(self):
        """
        Collection of common attributes that apply to this device
        """

    @property
    def device_id(self):
        """
        Device identifier (unique)
        """
        return self.item

    def load(self):
        """
        Load and construct the corresponding IDeviceController

        Note: this method can raise a whole bag of errors: ValidationError,
        PluginLookupError, PluginLoadError
        """
        device_ctrl_cls = self._loader.load_plugin(self.controller)
        return device_ctrl_cls(self)


class SoftwareConfiguration(Document):
    """
    Software configuration.

    Defines where to find the executable and how to run it
    """

    _loader = PluginLoader("lava.controllers.software", ISoftwareController)

    document_schema = {
        "type": "object",
        "default": {},
        "properties": {
            "executable": {
                "type": "string",
                "default": None,
                "optional": True
            },
            "environment": {
                "type": "object",
                "default": {},
                "optional": True,
                "additionalProperties": {
                    "type": "string"
                }
            }
        }
    }

    @bridge.readwrite
    def executable(self):
        """
        pathname of the executable

        Can be none, in that case the executable should be
        looked up in PATH.
        """

    @bridge.fragment
    def environment(self):
        """
        additional environment values

        Anything specified here is added to the existing environment
        when you start the executable.
        """

    @property
    def software_id(self):
        """
        Software identifier (unique)

        Abused to be the default name of the executable
        """
        return self.item

    def load(self):
        """
        Load and construct the corresponding ISoftware

        Note: when there is no specialized wrapper an instance of
        GenericSoftware is provided.

        Note: this method can raise a whole bag of errors: ValidationError,
        PluginLookupError, PluginLoadError
        """
        # XXX: self.item is the name of the key in the software dictionary, it
        # is not the same as executable as executable can include a full
        # pathname (for stuff that is not in PATH or stuff that is installed
        # with non-standard executable name)
        try:
            software_ctrl_cls = self._loader.load_plugin(self.item)
        except PluginLookupError:
            from lava.core.controllers.software import \
                    GenericSoftwareController
            software_ctrl_cls = GenericSoftwareController
        return software_ctrl_cls(
            self.value, self.item,
            self.document.credentials['sudo-password'].value)


# TODO: evolution support on incompatible changes, something we can take from
# linaro-dashboard-bundle. It would be nice if we could pull that from
# json_document. Note that it would _only_ be needed on _incompatible_ changes.
class Environment(Document):
    """
    Document describing the test environment.

    This is the single most important document in LAVA. It defines how the
    whole configuration looks like. Most of LAVA needs an instance of this
    to work.
    """

    document_schema = {
        "type": "object",
        "properties": {
            "name": {
                "type": "string"
            },
            "network": {
                "type": "object",
                "default": {},
                "optional": True,
                "properties": {
                    "proxy": {
                        "type": "object",
                        "default": {},
                        "optional": True,
                        "properties": {
                            "http": {
                                "type": "string",
                                "optional": True,
                                "default": None
                            },
                            "https": {
                                "type": "string",
                                "optional": True,
                                "default": None
                            }
                        }
                    }
                }
            },
            "credentials": {
                "type": "object",
                "default": {},
                "properties": {
                    "sudo-password": {
                        "type": "string",
                        "optional": True,
                        "default": None
                    }
                }
            },
            "software": {
                "type": "object",
                "default": {},
                "additionalProperties": SoftwareConfiguration
            },
            "devices": {
                "type": "object",
                "default": {},
                "additionalProperties": DeviceDescription
            }
        }
    }

    @bridge.readwrite
    def name(self):
        """
        Human readable name of the environment
        """

    @bridge.fragment
    def network(self):
        """
        Network settings
        """

    @bridge.fragment
    def devices(self):
        """
        Collection of devices
        """

    @bridge.fragment
    def credentials(self):
        """
        Collection of configuration for passwords and other secrets
        """

    @bridge.fragment
    def software(self):
        """
        Collection of configuration for external software
        """
