# Copyright (C) 2011-2012 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.formats
=================
"""

from json_document.document import Document
from json_document import bridge

__all__ = ['TestScenario']


class ScenarioAction(Document):
    """
    Fragment of the TestScenario document that describes a single action,
    or step, of the test scenario.
    """

    document_schema = {
        'type': 'object',
        'properties': {
            'command': {
                'type': 'string',
            },
            'parameters': {
                'optional': True,
                'type': 'object',
                'default': {},
            },
            'metadata': {
                'optional': True,
                'default': {},
            },
        },
        'additionalProperties': False
    }

    @bridge.readwrite
    def command(self):
        """
        Name of the action to perform

        Actions can be provided by any third party software. LAVA ships
        with a number of useful actions for both Ubuntu and Android.
        """

    @bridge.fragment
    def parameters(self):
        """
        Dictionary of parameters of this action
        """

    # XXX: this is ugly, why is it here?
    @bridge.readwrite
    def metadata(self):
        """
        Dictionary of meta-data
        """


class TestScenario(Document):
    """
    Document describing the test scenario to execute.

    The scenario is composed of the target definition and a list of actions to
    perform. Each action describes what to do (the action "command" name),
    along with any needed parameters.

    In addition to actions the scenario describes some additional things such
    as the device type, job name or health check flag. They are used for
    book-keeping in other parts of LAVA.
    """

    document_schema = {
        'type': 'object',
        'additionalProperties': {},
        'properties': {
            'actions': {
                'items': ScenarioAction
            }
        },
        'device_type': {
            'type': 'string',
            'optional': True,
            'default': None,
        },
        'job_name': {
            'type': 'string',
            'optional': True,
            'default': None
        },
        'health_check': {
            'optional': True,
            'default': False,
            'type': 'bool',
        },
        'target': {
            'type': 'string',
            'optional': True,
            'default': None,
        },
        'timeout': {
            'type': 'integer',
            'optional': False,
            'default': None,
        },
        'logging_level': {
            'type': 'string',
            'enum': ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"],
            'optional': True,
            'default': None
        }
    }

    @bridge.fragment
    def actions(self):
        """
        List of actions to perform as a part of this scenario
        """

    @bridge.readwrite
    def target(self):
        """
        Name of the device this test scenario is intended for
        """

    @bridge.readwrite
    def timeout(self):
        """
        Global job timeout in seconds
        """

    @bridge.readwrite
    def device_type(self):
        """"
        Type of the device this test scenario is intended for.
        """

    @bridge.readwrite
    def job_name(self):
        """
        Human-readable name of this job
        """

    @bridge.readwrite
    def health_check(self):
        """
        Flag used to mark health check jobs
        """
