# Copyright (C) 2011-2012 Linaro Limited
# vim: set fileencoding=utf8 :
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function

"""
lava.core.history
=================

Implementation of hierarhical history
"""

from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL
import collections
import datetime
import os
import sys
import tarfile
import traceback

from json_document.serializers import JSON
from json_schema_validator.extensions import (
    datetime_extension, timedelta_extension)
import simplejson


class DictProxy(object):

    def __init__(self, d):
        object.__setattr__(self, "_d", d)

    def as_dict(self):
        return object.__getattribute__(self, "_d")

    def __getattr__(self, name):
        try:
            return self.as_dict()[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        self.as_dict()[name] = value


ExceptionInfo = collections.namedtuple(
    "ExceptionInfo", "filename lineno function text")


_name_and_back = {
    DEBUG: "DEBUG",
    INFO: "INFO",
    WARNING: "WARNING",
    ERROR: "ERROR",
    CRITICAL: "CRITICAL",
    "DEBUG": DEBUG,
    "INFO": INFO,
    "WARNING": WARNING,
    "ERROR": ERROR,
    "CRITICAL": CRITICAL
}


class Activity(object):

    def __init__(self, level, message, **meta):
        """
        Create a new activity with the specified message, level and
        initial meta-data
        """
        # Message that has format() references to meta-data
        self._level = level
        self._message = message
        self._meta = DictProxy(meta)
        # File attachments
        self._attachments = []
        # Children activities
        self._sub_activities = []
        # Start and end timestamps
        self._start = datetime.datetime.utcnow()
        self._end = self._start
        # Exception + traceback
        self._exc_value = None
        self._exc_type = None
        self._traceback = None

    @property
    def raw_message(self):
        """
        the raw, unformatted message
        """
        return self._message

    @property
    def message(self):
        """
        rendered message

        Rendering substitutes str.format() style variables with data from
        self.meta.
        """
        return self._message.format(**self._meta.as_dict())

    @property
    def level(self):
        """
        level number, 10 (DEBUG) - 50 (CRITICAL)
        """
        return self._level

    @property
    def level_name(self):
        """
        name of the level "DEBUG" ... "CRITICAL"
        """
        return _name_and_back[self._level]

    @property
    def meta(self):
        """
        collection of meta-data

        Meta-data serves two purposes: to fill in data to the message
        template and to store arbitrary auxiliary data. This property
        returns a dictionary proxy that looks like an object (has normal
        attribute accessors) but still talks to the real dictionary behind
        the scenes. You can both read and write/set attributes at will.
        """
        return self._meta

    def attach(self, stream_or_pathname, name=None,  mime_type='text/plain'):
        """
        Add an attachment.

        Everything is extracted from the file-like object stream.  If you care
        about a particular name of the attachment it an be specified with name.
        MIME type can help the web front-end to render the content. It defaults
        to text/plain which is suitable for most log files.
        """
        if isinstance(stream_or_pathname, basestring):
            orig_name = stream_or_pathname
        else:
            orig_name = stream_or_pathname.name
        if name is None:
            name = orig_name
        self._attachments.append(
            DictProxy({
                'name': name,
                'orig_name': orig_name,
                'mime_type': mime_type}))

    @property
    def duration(self):
        """
        Duration of the activity.

        If the activity has not started the duration is always zero.
        Activities that are in progress will observe the duration to grow.
        """
        if self._start is None:
            return datetime.timedelta()
        if self._end is None:
            return datetime.datetime.utcnow() - self._start
        return self._end - self._start

    def as_json(self):
        """
        Convert this Activity to JSON
        """
        obj = simplejson.OrderedDict((
            ("level", self.level_name),
            ("message", self._message),
            ("meta", self.meta.as_dict()),
            ("start", (datetime_extension.to_json(self._start)
                       if self._start is not None else None)),
            ("duration", timedelta_extension.to_json(self.duration)),
            ("end", (datetime_extension.to_json(self._end)
                     if self._end is not None else None)),
            ("exc_value", self._exc_value),
            ("exc_type", (self._exc_type.__name__
                          if self._exc_type else None)),
            ("traceback", ([ExceptionInfo(*ei) for ei in
                            traceback.extract_tb(self._traceback)]
                           if self._traceback else None)),
            ("attachments", [
                attachment.as_dict() for attachment in self._attachments]),
            ("sub_activities", [
                activity.as_json() for activity in self._sub_activities])))
        return _build_json(obj)

    def __str__(self):
        """"
        Returns self.message
        """
        return self.message

    def save_as_tarball(self, scratch, tarball_stream, mode='w:gz'):
        """
        Save activity history as a tarball.

        The tarball will have two files 'history.txt' and 'history.json' with
        appropriate encoding of this activity, as well as all sub-activities.
        Any attachments are added as well.
        """
        with tarfile.open(tarball_stream.name, mode=mode,
                          fileobj=tarball_stream) as tarball:
            with scratch.open('history.txt') as stream:
                self.print_(stream=stream)
            tarball.add(stream.name, 'history.txt')
            with scratch.open('history.json') as stream:
                JSON.dump(stream, self.as_json())
            tarball.add(stream.name, 'history.json')
            self._save_tarball_attachments(tarball)

    def _save_tarball_attachments(self, tarball):
        for attachment in self._attachments:
            if os.path.isfile(attachment.orig_name):
                tarball.add(attachment.orig_name, attachment.name)
        for activity in self._sub_activities:
            activity._save_tarball_attachments(tarball)

    def __repr__(self):
        return "<Activity: %r>" % str(self)

    def __enter__(self):
        """
        Begins a 'long' activity.

        Long activities have both start and end, and thus, non-zero duration.
        This is called automatically by History.__enter__ so you won't usually
        have to use it explicitly.

        Here we reset self.end to None (it was initialized to 'now' in
        __init__)
        """
        self._end = None
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Terminates a 'long' activity.

        Here we set self.end to 'now' and record any exception information.
        See __enter__() for more information on 'long' activities.
        """
        self._end = datetime.datetime.utcnow()
        self._exc_value = exc_value
        self._exc_type = exc_type
        self._traceback = traceback

    def print_(self, stream=None, print_meta=False, print_children=True,
               stack=None):
        """
        Print a nice Unicode tree of activities.

        This is very useful for tracing and debugging.
        """
        if stack is None:
            stack = []
        if stream is None:
            stream = sys.stdout
        fragments = []
        if print_meta is False:
            fragments.append("{0:7}".format(self.level_name))
            fragments.append(" | ")
            fragments.append("{start}.{micro:06}".format(
                start=self._start.strftime("%Y-%m-%d %T"),
                micro=self._start.microsecond))
            fragments.append(" | ")
            if self.duration.total_seconds() > 0:
                fragments.append("{0:14}s".format(self.duration))
            else:
                fragments.append("{0:15}".format(""))
            fragments.append(" | ")
            fragments.append(''.join(self._tree_indent(stack)))
        # Process the message to make it safer
        message = self.message
        message = message.replace("\n", " ").replace("\r", " ")
        if isinstance(message, unicode):
            message = message.encode(stream.encoding or "UTF-8", 'replace')
        fragments.append(message)
        if self._exc_value:
            fragments.append(", ***crashed*** {0}".format(self._exc_value))
        print("".join(fragments), file=stream)
        if print_meta:
            print("\tstart: {0}".format(self._start))
            if self.duration.total_seconds() > 0:
                print("\tduration: {0}".format(self.duration))
            if self._end is not None and self._end != self._start:
                print("\tend: {0}".format(self._end))
            if self._exc_value is not None:
                print("\texc_value: {0}".format(self._exc_value))
            if self._exc_type is not None:
                print("\texc_type: {0}".format(self._exc_type))
            if self._traceback is not None:
                for index, exc_info in enumerate(self._traceback):
                    print("\ttraceback.{0}.filename: {1}".format(
                        index, self._traceback.filename))
                    print("\ttraceback.{0}.lineno: {1}".format(
                        index, self._traceback.lineno))
                    print("\ttraceback.{0}.function {1}".format(
                        index, self._traceback.function))
                    print("\ttraceback.{0}.text {1}".format(
                        index, self._traceback.text))
            for key, value in self.meta.as_dict().iteritems():
                # Skip one special key used by logging integration
                if key == "__prefmt_msg":
                    continue
                print("\t{key}: {value!r}".format(key=key, value=value))
        if print_children:
            pos = len(stack)
            stack.append(len(self._sub_activities))
            for sub in self._sub_activities:
                stack[pos] -= 1
                sub.print_(stream, print_meta, print_children, stack)
            stack.pop()

    @staticmethod
    def _tree_indent(stack, use_unicode=False):
        """
        Compute the indentation part of the tree displayed by print_.
        """
        # The code here might be tricky. It's easier to understand if you keep
        # in mind that stack[n] tells you how many more entries are left in the
        # tree at that stack depth. The stack is constructed by print_()
        for items_left in stack[:-1]:
            # FIXME: this is broekn
            if items_left == 0:
                yield "    "
            elif items_left == 1:
                if use_unicode:
                    yield "└   "
                else:
                    yield "\   "
            elif items_left > 1:
                if use_unicode:
                    yield "│   "
                else:
                    yield "|   "
        for items_left in stack[-1:]:
            if items_left == 0:
                if use_unicode:
                    yield "└── "
                else:
                    yield "\-- "
            elif items_left > 0:
                if use_unicode:
                    yield "├── "
                else:
                    yield "|-- "


def _build_json(obj):
    """
    Helper to construct json that keeps ordering and discards None values
    """
    if isinstance(obj, (dict, simplejson.OrderedDict)):
        new_obj = simplejson.OrderedDict()
        for key, value in obj.iteritems():
            if value is None or value == [] or value == {}:
                continue
            else:
                new_obj[str(key)] = _build_json(value)
        return new_obj
    elif isinstance(obj, list):
        return [_build_json(item) for item in obj]
    elif (isinstance(obj, (int, float, str, unicode, list, dict, bool))
          or obj is None):
        return obj
    else:
        try:
            return str(obj)
        except Exception:
            raise ValueError("cannot convert {0!r} to JSON".format(obj))


class History(object):
    """
    History class, provides helper methods for hierarchical logging.
    Maintains a stack of activities that are nested with context
    management.
    """

    def __init__(self, top):
        """
        Create a History that starts with the specified 'top' activity

        Usually you don't want to call this. Instead, use the singleton
        instance 'history'.
        """
        self._nesting = [top]
        self._last = None

    @property
    def top(self):
        """
        the top-level activity

        Very useful to grab entire history and do something with it, for
        example history.top.as_json() or history.top.print_()
        """
        return self._nesting[0]

    @property
    def bottom(self):
        """
        the bottommost activity

        This is the most nested activity, as established by:

            with history.xxx():
                ...
        """
        return self._nesting[-1]

    @property
    def last(self):
        """
        the most recent activity

        This is the most recently created activity in this history.
        """

    def log(self, level, message, **meta):
        """
        Create an activity and append it to the history
        """
        self._last = Activity(level, message, **meta)
        self.bottom._sub_activities.append(self._last)
        return self

    def debug(self, message, **meta):
        """
        Append DEBUG activity to the bottommost activity
        """
        return self.log(DEBUG, message, **meta)

    def info(self, message, **meta):
        """
        Append INFO activity to the bottommost activity
        """
        return self.log(INFO, message, **meta)

    def warning(self, message, **meta):
        """
        Append WARNING activity to the bottommost activity
        """
        return self.log(WARNING, message, **meta)

    def error(self, message, **meta):
        """
        Append ERROR activity to the bottommost activity
        """
        return self.log(ERROR, message, **meta)

    def critical(self, message, **meta):
        """
        Append CRITICAL activity to the bottommost activity
        """
        return self.log(CRITICAL, message, **meta)

    def __enter__(self):
        """
        Begins nested activity

        Nesting is widely useful as it helps to create a structure in the
        history stream. Typically you will call this method indirectly
        with code that looks like this:

            with history.info("Doing something complex"):
                history.debug("step 1")
                history.debug("step 2")
                history.debug("step 3")

        Here we also call __enter__() on the last activity to convert it to a
        'long' activity (one that has both start, end and non-zero duration)
        """
        self._nesting.append(self._last)
        return self._last.__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Terminates most recently nested activity

        Here we also call __exit__() on the last activity to allow it to record
        any exceptions that may have been raised.
        """
        last = self._nesting.pop()
        last.__exit__(exc_type, exc_value, traceback)


# Global history
history = History(Activity(DEBUG, "History module loaded"))
