# Copyright (C) 2011-2012 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.interfaces
====================

Interface classes used in LAVA
"""

import abc

from lava.core.utils import Interface


class ISoftwareController(Interface):
    """
    Controller for third party program invoked by LAVA
    """

    # XXX: sudo feels like an abuse, should have explicit 'sudo start'
    @abc.abstractmethod
    def __init__(self, configuration, executable, sudo_password):
        """
        Initialize software controller

        :param configuration: :class:`dict`:
            Configuration dictionary
        :param: executable: :class:`str`:
            Name of the executable to use. Note that this must not be
            the absolute pathname of the executable as that is looked
            up from PATH or explicitly defined in the configuration
        :param: sudo_password: :class:`str`:
            Password that should be feed to sudo if necessary. Can be
            None but this may cause some programs to fail.

        Initialize the controller with the specified configuration, executable
        pathname and optional sudo password. See
        :class:lava.core.software.GenericSoftware for the base implementation
        """

    @abc.abstractmethod
    def start(self, args, logfile=None, timeout=30, maxread=2000):
        """
        Run the external program with specified arguments.

        This is a thin abstraction around pexpect.spawn() but it also
        adds transparent support for hierarchical history and logging
        """

    @abc.abstractproperty
    def executable(self):
        """
        pathname to the executable of this program
        """

    @abc.abstractproperty
    def version(self):
        """
        version of the external software

        .. note::
            The value can be a long, multi line, string
        """


class IDeviceController(Interface):
    """
    Interface for device controllers

    Device controllers instructs LAVA how to configure the device and how to
    interpret test scenario actions. The actual class will typically implement
    additional interfaces.
    """

    @abc.abstractmethod
    def __init__(self, device_desc):
        """
        Initialize a device controller for a specific device.

        :param: device: :class:`lava.core.environment.DeviceDescription`:
            Device configuration from the environment object
        """

    @abc.abstractproperty
    def device_desc(self):
        """
        device part of the :class:`lava.core.environment.Environment`

        This is the same thing that was passed to __init__()
        """

    @abc.abstractproperty
    def configuration(self):
        """
        device configuration (as seen via controller specific configuration
        document class).

        To use this consult with the actual controller implementation.
        Usually it is only accessed by the controller itself during operation.
        """


class ITestScenarioInterpreter(Interface):
    """
    Something that can execute a test scenario

    Should be used as a mix-in interface for IDeviceController to indicate,
    that it can execute test scenarios.
    """

    @abc.abstractmethod
    def check_software(self, environment):
        """
        Check required third party software

        :param environment: :class:`lava.core.environment.Environment`:
            Environment object (used to locate software configuration)

        :raise :class:`lava.core.errors.PreconditionFailed`:
            When any of the required programs cannot be prepared

        This method should be called prior to calling execute_scenario() to
        ensure that required third party software is ready to use.
        """

    @abc.abstractmethod
    def emulate_scenario(self, scenario):
        """
        Emulate (dry-run) the test scenario

        :param scenario: :class:`lava.core.formats.TestScenario:
            The test scenario document

        :raise :class:`lava.core.errors.SemanticProblem`:
            The scenario cannot be executed for some reason

        This method can be used to run a whole scenario in 'dry run'
        mode where the interpreter checks how each step would alter the
        state of the test system. It should be used before really doing
        the scenario for real as it is very fast and does not allocate
        precious device resources.
        """

    @abc.abstractmethod
    def execute_scenario(self, scenario, session):
        """
        Execute a scenario for real

        :param scenario: :class:`lava.core.formats.TestScenario:
            The test scenario document

        :raise :class:`lava.core.errors.SemanticProblem`:
            The scenario has a semantic fault that would have been caught with
            emulate_scenario(). Under special circumstances emulate_scenario()
            can succeed while execute_scenario() can fail with this exception.
            It always indicates a bug in the implementation of one of the
            actions that have been executed thus far.
        :raise :class:`lava.core.errors.UnrecoverableProblem`:
            The scenario failed to execute for some reason

        After the scenario has finished `session.state` can be
        inspected for interesting output. In particular
        session.state['bundles'], if it exists, is a list of bundles
        that were produced during this session.
        """


class ITestScenarioAction(Interface):
    """
    Interface for all dispatcher actions.

    Actions are "steps" of a test scenario.
    """

    @abc.abstractmethod
    def __init__(self, parameters):
        """
        Initialize a action with the specified parameters

        :param parameters: :class:`dict`:
            Configuration dictionary. All action parameters
            must be stored there.

        :raise :class:`ValueError`:
            Usually raised as subclass (ValidationErrror) when
            parameters are incorrect. Note that parameters cannot be
            perfectly checked statically.
        """

    @abc.abstractmethod
    def check_precondition(self, state):
        """
        Check if state matches expectations

        :param state: :class:`dict`:
            Starting state to check against

        :raise :class:`lava.core.errors.SemanticProblem`:
            The action cannot be performed because of a combination
            of parameters and current state.

        This method can be used to spot actions that, while perfectly okay in
        other circumstances, make no sense having seen what has happened before
        (the state).

        It should be used before calling either emulate() or execute() to do
        most of the checking. It is somewhat weaker than them because it has no
        access to a device (which in turn makes it device independent and
        sufficient for earlier validation)
        """

    @abc.abstractmethod
    def emulate(self, controller, state):
        """
        Emulate the state change if this action was to be performed.

        :param controller: :class:`lava.core.interface.IDeviceController`:
            Device that this action operates against
        :param state: :class:`dict`:
            Temporary state (unlike session.state)

        :raise :class:`lava.core.errors.SemanticProblem`:
            The action cannot be performed because of a combination of
            parameters and current state.

        You can use this method, instead of calling perform(), to check perform
        is likely to fail. State is modified so make sure to pass separate
        state object to a sequence of emulate() calls and not reuse it with
        perform().

        .. note::
            You should call check_precondition() before calling emulate()
        """

    @abc.abstractmethod
    def execute(self, controller, session):
        """
        Execute this action in a specified session.

        :param controller: :class:`lava.core.interface.IDeviceController`:
            Device that this action operates against
        :param session: :class:`lava.core.session.Session`:
            Session for access to environment and scratch space

        :raise :class:`lava.core.errors.UnrecoverableProblem`:
            The action failed to perform for some reason
        :raise :class:`lava.core.errors.Timeout`:
            The action did not finish in time, it is a subclass of
            UnrecoverableProblem but it is stated here explicitly should the
            interpreter want to treat it specially.
        :raise :class:`lava.core.errors.SemanticProblem`:
            The action cannot be performed because of a combination of
            parameters and current state.

        Usually actions modify `session.state` to keep track of what they
        have done. Other commands reuse this information to drive their
        execution.

        .. note::
            You should call check_precondition() before calling execute()
        """


class IDeviceControllerMixIn(Interface):
    """
    Mix-in for :class`IDeviceController` classes.

    This base class is here to help identify device 'features' that
    actions can depend on. It has no actual methods or properties.

    Subclasses of this interface can define methods are so that actions
    can statically check if they can work with a specific device. This
    allows emulating the execution of an entire test scenario to check
    for issues before starting a lengthy process.
    """


class IBootable(IDeviceControllerMixIn):
    """
    Something that can be booted.

    Should be used as a mix-in interface for :class:`IDeviceController`.

    .. note::
        Those methods are placed in an interface so that actions can statically
        check if they can work with a specific device. This allows emulating
        the execution of an entire test scenario to check for issues before
        starting a lengthy process.
    """

    @abc.abstractmethod
    def boot(self, session, image, shell_prompt, timeout):
        """
        Boot the device with the specified image

        :param session: class:`lava.core.session.Session`:
            Session object used for scratch storage and software
            configuration.
        :param image:
            Image that we can boot. The type is not really specified
            and can differ from one situation to another.
        :param shell_prompt:
            Regular expression that describes the shell prompt after booting
        :param timeout:
            Timeout after which we know something has timed out

        :raise :class:`lava.core.errors.Timeout`:
            When timeout expires and we're still not ready
        :raise :class:`lava.core.errors.UnrecoverableProblem`:
            When something just goes wrong
        """

    @abc.abstractmethod
    def stop(self):
        """
        Stop the device, if running
        """


class IAgentHost(IDeviceControllerMixIn):
    """
    Something that can host an agent.

    Should be used as a mix-in interface for IDeviceController

    .. note::
        Those methods are placed in an interface so that actions can statically
        check if they can work with a specific device. This allows emulating
        the execution of an entire test scenario to check for issues before
        starting a lengthy process.
    """

    @abc.abstractmethod
    def start_agent(self):
        """
        Start LAVA agent on the device
        """
        # TODO: document exceptions (UnrecoverableProblem)

    @abc.abstractproperty
    def agent(self):
        """
        agent instance, if running
        """


# TODO IAgent interface
