# Copyright (C) 2011-2012 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.interpreter
=====================
"""

import abc

from lava.core.errors import PreconditionFailed, SemanticProblem
from lava.core.errors import UnrecoverableProblem
from lava.core.interfaces import ITestScenarioInterpreter


__all__ = ["TestScenarioInterpreterMixIn"]


class TestScenarioInterpreterMixIn(ITestScenarioInterpreter):
    """
    MixIn class for IDevice subclasses that want to support running tests
    """

    @abc.abstractmethod
    def get_action_map(self):
        """
        Return an dictionary mapping from method name to method implementation
        """

    # XXX: the list of required software is something that we should get out of
    # the emulation phase
    @abc.abstractmethod
    def get_required_software(self):
        """
        Return a list of executables that are needed
        """

    def check_software(self, environment):
        with self.history.info("Checking required external software"):
            for executable in self.get_required_software():
                software_conf = environment.software[executable]
                try:
                    software_ctrl = software_conf.load()
                except Exception as exc:
                    raise PreconditionFailed(
                        "Required program {0} is not available: {1}".format(
                            executable, exc))
                else:
                    self.history.info("Executable {executable} ready",
                                      executable=software_ctrl.executable)

    def emulate_scenario(self, scenario):
        action_map = self.get_action_map()
        # Start with empty state
        state = {}
        with self.history.info("Emulating test scenario"):
            # Walk over all the actions
            for index, action_def in enumerate(scenario.actions, 1):
                with self.history.info(
                    "Emulating action {index}: {command}",
                    index=index, command=action_def.command):
                    # Load the action
                    try:
                        action_cls = action_map[action_def.command]
                    except KeyError:
                        raise SemanticProblem(
                            ("Either command {0!r} has a typo or it cannot "
                             "be executed on this device").format(
                                 action_def.command))
                    try:
                        action = action_cls(action_def.parameters.value)
                    except ValueError as exc:
                        raise SemanticProblem(
                            ("Command {0!r} has incorrect parameters: "
                             "{1}").format(action_def.command, exc))
                    # Check if we can execute the action now
                    try:
                        action.check_precondition(state)
                    except SemanticProblem:
                        self.history.error(
                            """Precondition for action {command} has failed.
                            This almost almost always indicates that the
                            scenario is incorrect (as this failure happened
                            during emulation of the scenario, not real
                            execution. In rare cases it may also be a bug in
                            the emulate() function of any of the preceding
                            actions but those functions are usually very simple
                            and should not fail""", command=action_def.command)
                        raise
                    # Now pretend we had executed it
                    try:
                        action.emulate(self, state)
                    except SemanticProblem:
                        self.history.error(
                            """Emulation of action {command} has failed. This
                            almost always indicates a bug in LAVA please report
                            it with the exact output of lava --version and a
                            copy of your scenario.  See lava --help to find out
                            how to report bugs""", command=action_def.command)
                        raise
                    self.history.debug(
                        "State after emulation: {state!r}", state=state)

    def execute_scenario(self, scenario, session):
        action_map = self.get_action_map()
        with self.history.info("Executing test scenario"):
            for index, action_def in enumerate(scenario.actions, 1):
                with self.history.info(
                    "Executing action {index}: {command}",
                    index=index, command=action_def.command):
                    # Load the action
                    try:
                        action_cls = action_map[action_def.command]
                    except KeyError:
                        raise SemanticProblem(
                            ("Either command {0!r} has a typo or it cannot "
                             "be executed on this device").format(
                                 action_def.command))
                    try:
                        action = action_cls(action_def.parameters.value)
                    except ValueError as exc:
                        raise SemanticProblem(
                            ("Command {0!r} has incorrect parameters: "
                             "{1}").format(action_def.command, exc))
                    # Check if we can execute the action now
                    try:
                        action.check_precondition(session.state)
                    except SemanticProblem:
                        self.history.error(
                            """Precondition for action {command} has failed.
                            This may indicate a bug in the implementation of
                            any of the preceding actions as preconditions are
                            also checked by the scenario emulator. If you got
                            here then the one of the action implementations had
                            encountered a failure without raising
                            UnrecoverableError exception. This has allowed the
                            scenario to proceed until stopping here.""",
                            command=action_def.command)
                        raise
                    # Now really execute the action
                    try:
                        action.execute(self, session)
                    except UnrecoverableProblem:
                        self.history.error(
                            """Action {command} has failed to work. This is
                            not a problem with your scenario. It may indicate
                            that the action is implemented incorrectly or has
                            simply failed for environmental reasons (test
                            software not working, temporary network problems,
                            etc).""", command=action_def.command)
                        raise
                    self.history.debug("State after execution: {state!r}",
                                       state=session.state)
