# Copyright (C) 2011-2012 Linaro Limited
# vim: set fileencoding=utf8 :
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.logging
=================

Logging utilities
"""

import logging

from lava.core.history import history as global_history


class LoggerToHistoryProxy(object):
    """
    Wrapper that looks like logging.Logger but also creates history records
    """

    _history = global_history

    def __init__(self, logger):
        self._logger = logger

    def log(self, level, msg, *args, **kwargs):
        self._logger.log(level, msg, *args, **kwargs)
        self._history.log(level, "{__prefmt_msg}",
                     __prefmt_msg=msg % args, logger=self._logger.name)

    def debug(self, msg, *args, **kwargs):
        return self.log(logging.DEBUG, msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        return self.log(logging.INFO, msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        return self.log(logging.WARNING, msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        return self.log(logging.ERROR, msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        return self.log(logging.CRITICAL, msg, *args, **kwargs)

    def exception(self, msg, *args):
        self._logger.exception(msg, *args)


class HistoryToLoggerProxy(object):
    """
    Wrapper that looks like History but also logs to logger
    """

    _history = global_history

    def __init__(self, logger):
        self._logger = logger

    def log(self, level, message, **meta):
        self._logger.log(level, message.format(**meta))
        return self._history.log(level, message, **meta)

    def debug(self, message, **meta):
        return self.log(logging.DEBUG, message, **meta)

    def info(self, message, **meta):
        return self.log(logging.INFO, message, **meta)

    def warning(self, message, **meta):
        return self.log(logging.WARNING, message, **meta)

    def error(self, message, **meta):
        return self.log(logging.ERROR, message, **meta)

    def critical(self, message, **meta):
        return self.log(logging.CRITICAL, message, **meta)


class LoggingMixIn(object):
    """
    Mix-in that adds a self.logger and self.history instance attributes
    """

    @property
    def _logger(self):
        return logging.getLogger(
            self.__class__.__module__ + "." + self.__class__.__name__)

    @property
    def history(self):
        """
        magic lava.core.history bridge
        """
        return HistoryToLoggerProxy(self._logger)

    @property
    def logger(self):
        """
        magic classic python logging bridge
        """
        return LoggerToHistoryProxy(self._logger)
