# Copyright (C) 2010-2012 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function

"""
lava.core.main
==============

Implementation of the `lava-dev` shell command.

See the note in LavaDispatcher docstring about `lava-dev` vs `lava`
"""

import argparse
import logging
import os
import pdb
import sys

from lava.core.commands import Dispatcher
from lava.core.errors import CommandError
from lava.core.utils  import Scratch
from lava.core.history import history


class LavaDispatcher(Dispatcher):
    """
    Dispatcher implementing the `lava-dev` shell command (note that
    `lava-dev` shall be renamed to `lava` once lava-core replaces
    current group of lava-* projects.

    This dispatcher imports plugins from the pkg_resources namespace
    `lava-dev.commands`. Additional plugins can be registered as either
    :class:`lava.command.Command` or :class:`lava.command.CommandGroup`
    sub-classes.
    """

    name = "lava"
    help = None
    description = """
This executable allows you to work with LAVA. LAVA commands form a tree, where
you typically invoke sub or sub-sub commands to do actual things. Please
inspect the list below for a list below for details.
"""

    epilog = """
about LAVA
    LAVA, which stands for Linaro Automated Validation Architecture, is a
    project started by Linaro tasked with build a collection of freely
    available tools and equipment for test automation. LAVA is composed of many
    extensible pieces that together help to plan, schedule, execute, track, and
    run tests on virtual and physical devices, then finally to store and
    visualize the results.

    To learn more about LAVA please visit https://launchpad.net/lava-project/

about Linaro
    Linaro is a non-profit organization jointly formed by
    ARM, Canonical, TI, Freescale and Samsung to advance the state of free
    software on the ARM architecture.

    To learn more about Linaro please visit http://linaro.org/

software development and hacking:
    Source code is freely available and is managed by the Bazaar version
    control system and the Launchpad (http://launchpad.net/) collaboration
    platform. Note that while LAVA is located in more than one project you
    should eventually be able to look at the lava-core project for most of
    the code. The process of merging small lava projects into lava-core is
    not complete yet, as of May 2012.

    To get a copy of LAVA install bazaar and run the following command:

        $ bzr branch lp:lava-core

    We welcome all contributors, see HACKING to get started

    Development is discussed on the linaro-validation@lists.linaro.org mailing
    list, you can subscribe by visiting http://lists.linaro.org/

support:
    Report bugs at https://bugs.launchpad.net/lava-project/+filebug

    Ask questions in your native language at
    https://answers.launchpad.net/lava-project/+addquestion.

    You can also ask for real-time support in the #linaro-lava channel on the
    freenode IRC network. Keep in mind that although geographically distributed
    we (LAVA developers) may be away at a given time. Note that developer time
    is precious so please be considerate.

copyright and license:
    Copyright (c) Linaro Limited 2010-2012

    LAVA is free software, licensed under GLPv3, LGPLv3 and AGLPv3.
    Check the license header in a specific file to know more.
    """

    def __init__(self):
        # Call this early so that we don't get logging.basicConfig
        # being called by accident. Otherwise we'd have to
        # purge all loggers from the root logger and that sucks
        self.setup_logging()
        # Initialize the base dispatcher
        super(LavaDispatcher, self).__init__()
        # And import the non-flat namespace commands
        self.import_commands('lava_dev.commands')
        self.args = None
        self._history_needs_saving = True

    def construct_parser(self):
        """
        Construct a parser for this dispatcher.

        This is only used if the parser is not provided by the parent
        dispatcher instance.
        """
        # Construct a basic parser
        parser = super(LavaDispatcher, self).construct_parser()
        # Add the --verbose flag
        parser.add_argument(
            "-v", "--verbose",
            default=False,
            action="store_true",
            help="be more verbose (displays all INFO messages)")
        group = parser.add_argument_group(title="history management")
        group.add_argument(
            "-H", "--history",
            type=argparse.FileType(mode="wb"),
            metavar="TARBALL",
            help="Save session history, as tarball, to TARBALL")
        group = parser.add_argument_group(
            title="debugging and software development")
        # Add the --debug flag
        group.add_argument(
            "-D", "--debug",
            action="store_true",
            default=False,
            help="enable DEBUG messages on the root logger")
        # Add the --trace flag
        group.add_argument(
            "-T", "--trace",
            metavar="LOGGER",
            action="append",
            default=[],
            help=("enable DEBUG messages on the specified logger "
                  "(can be used multiple times)"))
        # Add the --pdb flag
        group.add_argument(
            "-P", "--pdb",
            action="store_true",
            default=False,
            help="jump into pdb (python debugger) when a command crashes")
        group.add_argument(
            "-C", "--debug-command-errors",
            action="store_true",
            default=False,
            help="crash on CommandError, useful with --pdb")
        # Add the --debug-interrupt flag
        group.add_argument(
            "-I", "--debug-interrupt",
            action="store_true",
            default=False,
            help="crash on SIGINT/KeyboardInterrupt, useful with --pdb")
        # Return the improved parser
        return parser

    def setup_logging(self):
        """
        Setup logging for the root dispatcher
        """
        self._setup_error_handler()
        self._setup_warning_handler()

    def dispatch(self, raw_args=None):
        """
        Enhanced version of Dispatcher.dispatch() that logs command duration
        and handles all exceptions (swallowing them)
        """
        try:
            return super(LavaDispatcher, self).dispatch(raw_args, True)
        except SystemExit:
            pass
        except BaseException as exc:
            if isinstance(exc, KeyboardInterrupt):
                self.say(None, "Interrupted!")
                if self.args is None or not self.args.debug_interrupt:
                    return 1
            elif isinstance(exc, CommandError):
                if self.args is None or not self.args.debug_command_errors:
                    return 1
            else:
                self.history.error(
                    "Runaway exception while invoking command: {exc}",
                    exc=str(exc))
                # TODO: Add explicit traceback logger so that full backtrace is
                # not displayed directly
                self.logger.exception("Poof!")
            if self.args and self.args.pdb:
                self.logger.info("Starting pdb...")
                pdb.post_mortem()
            else:
                self.say(None, "This command has crashed")
                self.say(None, "The problem was: {0}", exc)
                self.say(None, "See {0} for full traceback",
                         self.error_log_pathname),
            return 1
        finally:
            # Save history as tarball
            if self._history_needs_saving and self.args and self.args.history:
                self.history.debug("Saving history to {name}",
                                   name=self.args.history.name)
                with Scratch() as scratch:
                    self.save_history_now(scratch)

    def save_history_now(self, scratch):
        self._history_needs_saving = False
        if self.args is not None and self.args.history:
            history.top.save_as_tarball(scratch, self.args.history)

    def _setup_error_handler(self):
        """
        Add a handler that logs all ERRORs to a dedicated file
        """
        # TODO: support XDA dirs
        pathname = os.path.expanduser("~/.cache/lava/logs/error.log")
        self.error_log_pathname = pathname
        if not os.path.exists(os.path.dirname(pathname)):
            os.makedirs(os.path.dirname(pathname))
        handler = logging.FileHandler(pathname, delay=True)
        handler.setLevel(logging.ERROR)
        handler.setFormatter(
            logging.Formatter("%(levelname)s %(name)s: %(message)s"))
        logging.getLogger().addHandler(handler)

    def _setup_debug_handler(self):
        """
        Add a handler that logs everything to a dedicated file
        """
        # TODO: support XDA dirs
        pathname = os.path.expanduser("~/.cache/lava/logs/debug.log")
        self.say(None, "Note: debugging messages are saved to {0}", pathname)
        if not os.path.exists(os.path.dirname(pathname)):
            os.makedirs(os.path.dirname(pathname))
        handler = logging.FileHandler(pathname)
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(
            logging.Formatter((
                "%(asctime)s "
                "[pid:%(process)s, thread:%(threadName)s, "
                "reltime:%(relativeCreated)dms] "
                "%(levelname)s %(name)s: %(message)s")))
        logging.getLogger().addHandler(handler)

    def _setup_warning_handler(self):
        """
        Add a handler that logs warnings, and other important stuff to stderr
        """
        handler = logging.StreamHandler(sys.stderr)
        handler.setLevel(logging.WARN)
        handler.setFormatter(
            logging.Formatter("%(levelname)s: %(name)s: %(message)s"))
        logging.getLogger().addHandler(handler)

    def _setup_info_handler(self):
        class OnlyInfoFilter(logging.Filterer):
            def filter(self, record):
                if record.levelno == logging.INFO:
                    return 1
                return 0

        class SayHandler(logging.Handler):
            def emit(inner_self, record):
                self.say(None, "{0}", record.getMessage())
        handler = SayHandler()
        handler.setLevel(logging.INFO)
        handler.setFormatter(logging.Formatter("%(message)s"))
        handler.addFilter(OnlyInfoFilter())
        logging.getLogger().addHandler(handler)

    def _adjust_logging_level(self, args):
        # XXX: store arguments as we want to look at some more esoteric options
        # in dispatch() later on, this feels ugly, perhaps the dispatcher
        # should just know the arguments
        self.args = args
        # Enable verbose message handler
        if args.verbose:
            logging.getLogger().setLevel(logging.INFO)
            self._setup_info_handler()
        # Enable debug.log if needed
        if args.debug or args.trace:
            self._setup_debug_handler()
        # Enable application-wide debugging
        if args.debug:
            logging.getLogger().setLevel(logging.DEBUG)
        # Enable trace loggers
        for name in args.trace:
            logging.getLogger(name).setLevel(logging.DEBUG)
