# Copyright (C) 2011-2012 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.plugins
=================

Formalized plugin system for LAVA.
"""

import pkg_resources
from simplejson import OrderedDict

from lava.core.errors import PluginLookupError, PluginLoadError
from lava.core.logging import LoggingMixIn


class PluginLoader(LoggingMixIn):
    """
    Loader for plugins that populate a specified namespace and conform to a
    specified interface. It is expected that the objects pointed to by the
    entry point system are _classes_. The plugin loader will call issubclass()
    on them.

    This plugin loader is based on pkg_resources machinery but it may be
    expected to behave specially in the future. For example, when interacting
    from a frozen binary on windows.

    .. note::
        Each instance has a cache of entry points that were discovered in the
        past that is never purged. This should not be a problem under normal
        circumstances.
    """

    def __init__(self, namespace, interface):
        """
        Create a plugin loader.
        """
        self.namespace = namespace
        self.interface = interface
        self._entry_points = None  # Cache of entry_point.name -> entry_point

    def load_all_plugins(self):
        """
        Load all plugins, returns an dict of objects.
        Plugins that fail to load are skipped.
        """
        self._iter_entry_points_if_needed()
        plugins = OrderedDict()
        for entry_point in self._entry_points.itervalues():
            try:
                plugin = self._load(entry_point)
            except PluginLoadError:
                pass
            else:
                plugins[entry_point.name] = plugin
        return plugins

    def load_plugin(self, name):
        """
        Load a plugin by name.

        @raises PluginLookupError if the plugin cannot be found.
        @raises PluginLoadError if the plugin cannot be loaded
        """
        self._iter_entry_points_if_needed()
        try:
            entry_point = self._entry_points[name]
        except KeyError:
            raise PluginLookupError(name)
        else:
            return self._load(entry_point)

    def _load(self, entry_point):
        """
        Load an validate an entry point object.

        ImportErrors and DistributionNotFound errors are caught and trigger a
        diagnostic message, the return value is None if that happens.

        The value loaded from the entry point must be a class that is a
        subclass of the interface class passed to PluginLoader constructor.
        Failure to conform triggers another diagnostic message and a None
        return value.
        """
        try:
            self.logger.debug("Loading entry point: %r", entry_point)
            plugin = entry_point.load()
        except (ImportError, pkg_resources.DistributionNotFound):
            self.logger.exception(
                "Unable to load plugin: %s (%r)",
                entry_point.name, entry_point)
            raise PluginLoadError(entry_point)
        else:
            if (not isinstance(plugin, type)
                or not issubclass(plugin, self.interface)):
                self.logger.exception(
                    "Plugin %r does not implement interface %r",
                    entry_point.name, self.interface)
                raise PluginLoadError(entry_point)
            return plugin

    def _iter_entry_points_if_needed(self):
        """
        Iterate all the entry points in a specified namespace.

        This method keeps the result in self._entry_points because
        pkg_resources.iter_entry_points() is surprisingly slow, doing a lot of
        IO on tiny files all over the disk.
        """
        if self._entry_points is None:
            self._entry_points = OrderedDict()
            for entry_point in pkg_resources.iter_entry_points(self.namespace):
                self._entry_points[entry_point.name] = entry_point
