# Copyright (C) 2011-2012 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.utils
===============

Common utilities:

    Scrach - temporary filesystem space
    Network - proxy management and easy downloads
    Download - complex downloads
    ObjectWithConfiguration - json-document + code
"""

import abc
import datetime
import os
import shutil
import sys
import tempfile

from json_document.document import Document
import requests
import requests.exceptions

from lava.core.errors import UnrecoverableProblem
from lava.core.logging import LoggingMixIn


class Scratch(LoggingMixIn):
    """
    Scratch space.

    Creates and manages a temporary directory. The .open() helper creates a
    temporary file that is ready to go and will be removed once the scratch
    space goes away.
    """

    def __init__(self, pathname=None, suffix='.scratch', prefix='tmp'):
        """
        Create a scratch space.
        """
        self._dir = tempfile.mkdtemp(suffix, prefix, dir=pathname)
        self.logger.debug("Created scratch directory: %r", self._dir)

    def cleanup(self):
        if os.path.exists(self._dir) and not os.environ.get("RETAIN_SCRATCH"):
            self.logger.debug("Removing scratch directory: %r", self._dir)
            shutil.rmtree(self._dir, ignore_errors=True)

    def __enter__(self):
        return self

    def __exit__(self, exc_value, exc_type, traceback):
        self.cleanup()

    def __del__(self):
        self.cleanup()

    def just_name(self, name):
        """
        Return a pathname to a temporary location with the specified name.

        This function is useful when you need to interact with external
        software that will create files or directories but needs a pathname to
        operate
        """
        return os.path.join(self._dir, name)

    def open(self, name, mode='w+b'):
        """
        Open a file with the specified name
        """
        return open(os.path.join(self._dir, name), mode)

    def unique(self, suffix='', prefix='tmp.', mode='w+b'):
        """
        Create a temporary file with unique name.

        This file will be deleted once the scratch space is discarded.
        The name of the file can be accessed with the .name property.
        """
        return tempfile.NamedTemporaryFile(mode, suffix=suffix, prefix=prefix,
                                           dir=self._dir, delete=False)


class TextDownloadProgress(object):
    """
    Helper that tracks progress of an ongoing Download
    """

    def __init__(self, url, size=None, update_interval=1.0):
        self._url = url
        self._bytes = 0  # bytes so far
        self._size = size  # total size, None if unknown
        self._start = None
        self._finish = None
        self._update_interval = update_interval

    @property
    def percentage(self):
        """
        Percentage of the download so far
        """
        if self._size is None:
            return None
        else:
            return self._bytes * 100.0 / self._size

    def get_message(self):
        if self._size and self._size == self._bytes:
            return "Downloaded {0} ({1}MB in {2})".format(
                os.path.basename(self._url),
                self._bytes >> 20, self._finish -
                self._start)
        if self._size:
            return "Downloading {0} (total {1}MB, {2:02.1f}%)".format(
                os.path.basename(self._url), self._size >> 20,
                self.percentage)
        else:
            return "Downloading {0} (total ???, {1}MB so far)".format(
                os.path.basename(self._url), self._bytes >> 20)

    def _begin(self):
        self._start = datetime.datetime.utcnow()
        self._last_update = self._start

    def _chunk(self, chunk):
        self._bytes += len(chunk)
        now = datetime.datetime.utcnow()
        if (now - self._last_update).total_seconds() < self._update_interval:
            return
        self._last_update = now
        print "\r%s" % self.get_message(),
        sys.stdout.flush()

    def _end(self):
        self._finish = datetime.datetime.utcnow()
        print "\r%s" % self.get_message()
        sys.stdout.flush()


class Download(LoggingMixIn):
    """
    Helper for downloading things.

    Minimalistic usage:

        Download(url).save(stream)

    More elaborate usage:

        Download(url, progress=TextDownloadProgress).save(stream)
    """

    def __init__(self, url, proxies=None, progress_cls=None):
        """
        Start downloading the specified resource

        Raises an exception on error (yeah, helpful)
        """
        self.url = url
        with self.history.debug("Connecting to {url}", url=url,
                                proxies=proxies):
            self._response = requests.get(url, proxies=proxies)
            # If we did not get a 200, abort
            self._response.raise_for_status()
        content_length = int(self._response.headers.get(
            'content-length', None))
        self._progress = progress_cls(url, content_length)

    def __repr__(self):
        return "<Download url:%r>" % self.url

    def save(self, stream):
        """
        Save the payload to a stream
        """
        with self.history.debug("Saving data to {stream}", stream=stream.name):
            if self._progress:
                self._progress._begin()
            for chunk in self._response.iter_content(chunk_size=409600):
                if self._progress:
                    self._progress._chunk(chunk)
                stream.write(chunk)
            if self._progress:
                self._progress._end()


class Network(LoggingMixIn):
    """
    Network support, used for downloading files.
    """

    def __init__(self, scratch, http_proxy=None, https_proxy=None):
        self._scratch = scratch
        self._proxies = {'http': http_proxy, 'https': https_proxy}

    def download(self, url, hint=''):
        """
        Download something to temporary session storage

        Extra suffix, if provided, helps to identify the file as if will be
        used in the part of the temporary name.

        @return Path of the downloaded file
        @raises ValueError when the URL cannot be retrieved,
                requests.* for other issues
        """
        with self.history.debug("Downloading {url}", url=url):
            try:
                down = Download(url, self._proxies, TextDownloadProgress)
            except requests.exceptions.RequestException as exc:
                raise UnrecoverableProblem(
                    "Cannot download: {0}: {1}".format(url, exc))
            with self._scratch.unique(prefix='download.',
                                      suffix="." + hint) as stream:
                down.save(stream)
            return stream.name


class ObjectWithConfiguration(object):
    """
    Object with a JSON document-based onfiguration.

    Subclasses should override the default Configuration class
    """

    class Configuration(Document):
        """
        Document describing configuration of this class
        """

    def __init__(self, config):
        self._configuration = self.Configuration(config)
        self._configuration.validate()

    @property
    def configuration(self):
        return self._configuration


class Decoy(object):
    """
    Dummy value that has a description

    Useful instead of an object()
    """

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "Decoy({0!r})".format(self.value)


class Interface(object):
    """
    Interface class.

    This trivial class simply wraps the ABCMeta metaclass so you don't have to
    worry about it much.
    """

    __metaclass__ = abc.ABCMeta
