# Copyright (C) 2011-2012 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.workspace
===================
"""

import os

from json_document.shortcuts import persistence

from lava.core.environment import Environment
from lava.core.session import Session
from lava.core.logging import LoggingMixIn


class Workspace(LoggingMixIn):
    """
    Container for LAVA configuration and other runtime data. Also know as the
    .lava directory
    """

    def __init__(self, location):
        """
        Initialize a workspace with an existing on-disk location.  location
        must point to a directory that has the '.lava' sub-directory.
        """
        self._location = os.path.abspath(location)
        self._environment = Environment({})
        self._env_persistence = persistence(
            self._environment, self._environment_pathname)

    @property
    def location(self):
        """
        location of this workspace

        The location does not include the .lava directory
        """
        return self._location

    @property
    def environment(self):
        """
        :class:`lava.core.environment.Environment` object of this workspace

        .. note::
            This object is initially empty (empty environment), to use an
            actual environment you must call load_environment() and deal with
            any errors.
        """
        return self._environment

    def load_environment(self):
        """
        Load and return the workspace environment
        """
        # TODO: fs lock
        self._env_persistence.load()
        return self._environment

    def save_environment(self):
        """
        Save environment back to the filesystem, if needed
        """
        # TODO: fs lock
        self._env_persistence.save()

    @classmethod
    def create(cls, location='.'):
        """
        Create a new workspace in the specified location.

        Parent workspace lookup is _not_ performed. This is a filesystem
        synchronization/race point, if the caller looses an OSError is raised.
        """
        os.makedirs(os.path.join(location, '.lava'))
        self = cls(location)
        with open(self._environment_pathname, 'wt') as stream:
            stream.write("{ }")
        return self

    @classmethod
    def find(cls, start_dir='.'):
        """
        Find a Workspace, starting the search at start_dir
        """
        location = os.path.abspath(start_dir)
        while location and location != '/':
            candidate = os.path.join(location, '.lava')
            if os.path.isdir(candidate):
                return cls(location)
            location, junk = os.path.split(location)
        raise LookupError(".lava directory not found")

    def session(self):
        """
        Create and return a new Session object
        """
        return Session(self._environment)

    @property
    def _lava_dir(self):
        """
        pathname of the .lava directory
        """
        return os.path.join(self._location, '.lava')

    @property
    def _environment_pathname(self):
        """
        pathname of the 'environment.json' file
        """
        return os.path.join(self._lava_dir, 'environment.json')
