# Copyright (C) 2011-2012 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.agents
======================

Implementation of various device agents
"""

import shlex
import pkg_resources

from lava.core.logging import LoggingMixIn
from lava.core.errors import AgentError


class ShellCodeAgent(LoggingMixIn):
    """
    A simple agent implemented in plain POSIX shell.

    Once initialized, the shell run on the other end (virtual machine or board)
    and interprets our commands.  Since we can control it we don't need that
    much configuration do do what we need (as all the output is known).

    The agent is implemented as a shell function `lava_agent`. This function
    displays a standard prompt and loops on commands. The commands are fairly
    basic, run a command, read remote file, write remote file, stop (exit the
    agent).
    """

    SHELLCODE = """
    """

    def __init__(self, proc):
        """
        Create and install the shell agent connected to the pexpect object
        `proc`
        """
        self._proc = proc
        self._install()
        self._sequence = 0

    @property
    def sequence(self):
        """
        Sequence number for commands.

        This number is incremented with each command invoked via the agent
        """
        return self._sequence

    def invoke(self, command, timeout=None):
        """
        Run a command on the device.

        The output of the command is intercepted and is not available
        immediately. You can call return_code() output() and error() to read
        the return code, standard output and standard error respectively.
        """
        self._sequence += 1  # NOTE: we pre-increment, just like agent.sh
        with self.history.debug(("Invoking command {sequence} via the agent:"
                                 " {command}"), sequence=self._sequence,
                                command=command):
            # XXX: list is needed to run the generator that _cmd() creates
            list(self._cmd("invoke", command, timeout))

    def return_code(self):
        """
        Return the return code of the last command invoked with invoke()
        """
        self.logger.debug("Transferring return code of the last command")
        text = ''
        for chunk in self._cmd("return-code"):
            text += chunk.decode("ASCII", "replace")
        return int(text)

    def stdout(self):
        """
        Read the standard output of the last command
        """
        with self.history.debug(
            "Transferring standard output of the last command"):
            return self._cmd("stdout")

    def stderr(self):
        """
        Read the standard error of the last command
        """
        with self.history.debug(
            "Transferring standard error of the last command"):
            return self._cmd("stderr")

    def read(self, pathname):
        """
        Read an arbitrary file,

        Returns a generator of byte fragments
        """
        with self.history.debug("Transferring file {pathname}",
                                pathname=pathname):
            return self._cmd("read", pathname)

    def exit(self):
        """
        Terminate the agent
        """
        with self.history.debug("Terminating the agent"):
            self._cmd("exit")

    def _cmd(self, command, parameter=None, timeout=None):
        """
        Run an agent command.

        The response may contain a 'data' section of arbitrary size. That
        data is created with 'od -t x1'.

        Typical output looks like this::

            $ echo "LAVA is nice" | od -t x1
            0000000 4c 41 56 41 20 69 73 20 6e 69 63 65 0a
            0000015

        We just strip the first 8 bytes and decode the rest. Since the output
        may be random 8bit gibberish we try to do our best to convert it to an
        Unicode string. Each chunk is decoded separately as ASCII, any
        decoding errors are replaced with the funny Unicode question mark
        symbol.

        The output is streamed as it becomes available.
        """
        with self.history.debug("Waiting for agent prompt"):
            self._proc.sendline("")
            self._proc.expect_exact("lava-agent:")
        self.history.debug("Sending agent command: {command}", command=command)
        self._proc.sendline(command)
        if parameter is not None:
            with self.history.debug("Waiting for parameter prompt"):
                self._proc.expect_exact("parameter:")
            self.history.debug("Sending command parameter: {parameter}",
                               parameter=parameter)
            self._proc.sendline(parameter)
        with self.history.debug(("Waiting for the command to finish (timeout:"
                                 " {timeout})"), timeout=timeout):
            match = self._proc.expect_exact(
                ["ok", "data", "error"], timeout=timeout)
        self.history.debug("Agent command finished: {response}",
                           response=['ok', 'data', 'error'][match])
        if match == 1:
            # Read and process the data
            with self.history.debug("Reading escaped binary data..."):
                while True:
                    line = self._proc.readline().strip()
                    if line == "ok":
                        break
                    # self.logger.debug("Got encoded data: %r", line)
                    if len(line) > 8:
                        line = line[8:]
                        data = bytearray([int(val, 16) for val in
                                          line.split()])
                        # self.logger.debug("Decoded: %r", data)
                        yield data
        elif match == 2:
            raise AgentError("Agent request failed: %s" % command)

    @staticmethod
    def _get_inline_shellcode():
        """
        Get the agent shellcode as one string that can be pasted to a shell
        """
        shellcode = pkg_resources.resource_string(
            pkg_resources.Requirement.parse("lava-core"),
            'lava/core/agent.sh')
        # comments=True strips out comments
        return ' '.join(shlex.split(shellcode, comments=True))

    def _install(self):
        """
        Install the agent on the device that we can talk to over self._proc
        pexpect session.
        """
        # Stop the echo, we don't want to confuse the shellcode with the
        # patterns we expect below
        # self.logger.debug("Disabling echo")
        # self._proc.setecho(False)
        # Install our agent on the device.
        inline_shellcode = self._get_inline_shellcode()
        with self.history.debug("Injecting agent shell code: {shellcode!r}",
                                shellcode=inline_shellcode):
            # Do it slowly, one at a time
            for char in inline_shellcode:
                self._proc.write(char)
            self._proc.sendline()
        # And make sure it's ready
        with self.history.debug("Waiting for the agent to respond"):
            self._proc.expect_exact("lava-agent ready", timeout=600)
