# Copyright (C) 2011-2012 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.controllers.devices.qemu
==================================

Implementation of the qemu device controller.
"""

from json_document import bridge
from json_document.document import Document

from lava.core.actions.lava import lava_test_install, lava_test_run
from lava.core.actions.linaro import boot_linaro_image
from lava.core.agents import ShellCodeAgent
from lava.core.errors import SemanticProblem, PreconditionFailed
from lava.core.interfaces import IDeviceController, IAgentHost, IBootable
from lava.core.interfaces import ITestScenarioAction
from lava.core.interpreter import TestScenarioInterpreterMixIn
from lava.core.logging import LoggingMixIn
from lava.core.utils import ObjectWithConfiguration, Decoy


__all__ = ["QemuDeviceController"]


class deploy_linaro_image(LoggingMixIn, ObjectWithConfiguration,
                          ITestScenarioAction):
    """
    Deploy a typical Linaro image composed of a hardware pack and root
    filesystem. Both parts can be downloaded and composed together with
    linaro-media-create. Alternatively a direct URL to a pre-made image can be
    used.

    This action is only intended for the QemuDeviceController class.
    """

    class Configuration(Document):

        # This is how the schema for parameters should look, but there are bugs
        # in json_schema_validation that means it doesn't work (see
        # https://github.com/zyga/json-schema-validator/pull/6).
        document_schema = {
            'type': [
                {
                    'type': 'object',
                    'properties': {
                        'image': {
                            'type': 'string'
                        }
                    },
                    'additionalProperties': False,
                },
                {
                    'type': 'object',
                    'properties': {
                        'hwpack': {
                            'type': 'string'
                        },
                        'rootfs': {
                            'type': 'string'
                        },
                        'kernel_matrix': {
                            'type': 'string',
                            'optional': True
                        },
                        'use_cache': {
                            'type': 'bool',
                            'optional': True,
                            'default': True
                        },
                        'rootfstype': {
                            'type': 'string',
                            'optional': True,
                            'default': 'ext3'
                        },
                    },
                    'additionalProperties': False
                }
            ]
        }

        # TODO: remove the temporary, suboptimal schema once
        # json-schema-validator is fixed
        document_schema = {
            'type': 'object',
            'properties': {
                'hwpack': {
                    'type': 'string',
                    'optional': True,
                    'default': None
                },
                'rootfs': {
                    'type': 'string',
                    'optional': True,
                    'default': None
                },
                'image': {
                    'type': 'string',
                    'optional': True,
                    'default': None
                },
                # Unsupported, ignored
                'kernel_matrix': {
                    'type': 'string',
                    'optional': True
                },
                # Deprecated, ignored
                'use_cache': {
                    'type': 'bool',
                    'optional': True
                },
                'rootfstype': {
                    'type': 'string',
                    'optional': True,
                    "default": "ext4"
                },
                "image_size": {
                    "type": "string",
                    "optional": True,
                    "default": "3G"
                    # TODO: add regexp validation
                }
            },
            'additionalProperties': False,
        }

        @bridge.readonly
        def image(self):
            """
            URL of the pre-built image to deploy
            """

        @bridge.readonly
        def rootfs(self):
            """
            URL of the linaro rootfs to deploy
            """

        @bridge.readonly
        def hwpack(self):
            """
            URL of the linaro hwpack to deploy
            """

        @bridge.readonly
        def rootfstype(self):
            """
            Type of the root filesystem to use
            """

        @bridge.readonly
        def image_size(self):
            """
            Size of the combined image

            Typically this is an expression such as 3G for a three gigabyte
            image.
            """

    def check_precondition(self, state):
        # Check if we are not being run twice
        if 'image' in state:
            raise PreconditionFailed(
                "image is already deployed, deploying twice?")
        # Check if the argument combination makes sense
        if 'hwpack' in self.configuration:
            if 'rootfs' not in self.configuration:
                raise SemanticProblem(
                    'must specify rootfs when specifying hwpack')
            if 'image' in self.configuration:
                raise SemanticProblem(
                    'cannot specify image and hwpack')
        elif 'image' not in self.configuration:
            raise SemanticProblem(
                'must specify image if not specifying a hwpack')
        elif 'kernel_matrix' in self.configuration:
            raise SemanticProblem(
                'cannot specify kernel_matrix with an image')

    def emulate(self, controller, state):
        if not isinstance(controller, QemuDeviceController):
            raise SemanticProblem("This action is only designed for qemu")
        # Pretend we made an image
        state['image'] = Decoy("image to boot")

    def execute(self, controller, session):
        # Prepare the image
        if self.configuration.image:
            image = self._use_prebuilt_image(session)
        else:
            image = self._use_partial_image(session)
        # Remember the image to boot
        session.state['image'] = image

    def _use_prebuilt_image(self, session):
        # TODO: wrap download issues as UnrecoverableProblem
        return session.network.download(
            self.configuration.image, hint='image')

    def _use_partial_image(self, session):
        # Download both pieces
        # TODO: wrap download issues as UnrecoverableProblem
        rootfs = session.network.download(
            self.configuration.rootfs, hint='rootfs')
        hwpack = session.network.download(
            self.configuration.hwpack, hint="hwpack")
        # And put them together
        image_name = session.scratch.just_name('test.img')
        with session.scratch.unique(prefix="linaro-media-create.",
                                    suffix=".log") as stream:
            with self.history.info(("Using linaro-media-create to generate"
                                    "  bootable image...")):
                # Load linaro-media-create
                software = session.environment.software
                lmc = software["linaro-media-create"].load()
                # TODO: catch and wrap problems as Timeout and
                # UnrecoverableProblem, maybe do that in generate_image()
                # itself.
                lmc.generate_image([
                    "--dev", "beagle",  # for qemu we always use beagle
                    "--image-file", image_name,
                    "--binary", rootfs,
                    "--hwpack", hwpack,
                    '--hwpack-force-yes',  # install unverified hwpacks
                    "--rootfs", self.configuration.rootfstype,
                    "--image-size", self.configuration.image_size],
                    logfile=stream)
        return image_name


class QemuDeviceController(ObjectWithConfiguration,
                           LoggingMixIn,
                           TestScenarioInterpreterMixIn,
                           IDeviceController, IAgentHost, IBootable):
    """
    Controller for virtual devices implemented with qemu-system-*
    """

    class Configuration(Document):
        """
        Configuration for the Qemu device controller.

        The actual values are supplied by the device 'configuration' from the
        test environment. They are populated when you instantiate
        QemuDeviceController.
        """

        document_schema = {
            'type': 'object',
            'properties': {
                'qemu_system': {
                    'type': 'string',
                    'enum': ['qemu-system-arm']
                },
                'qemu_machine': {
                    'type': 'string',
                    'enum': ['beaglexm']
                }
            },
            'additionalProperties': False
        }

        @bridge.readonly
        def qemu_system(self):
            """
            Name of the qemu-system-xxx executable to use.

            NOTE: this should not be the full executable path. The path is
            defined in environment software. Typically, for arm, we want to use
            qemu-system-arm.
            """

        @bridge.readonly
        def qemu_machine(self):
            """
            Name of the qemu machine name to use.

            This corresponds to -M option for qemu-system-*
            """

    # IDeviceController methods

    def __init__(self, device_desc):
        self._agent = None   # ShellCodeAgent
        self._device_desc = device_desc  # DocumentFragment with device data
        self._log_file = None  # Qemu log
        self._proc = None  # Qemu process (pexpect.spawn)
        # NOTE: super-call __init__ late as we may crash and call __del__
        ObjectWithConfiguration.__init__(self, device_desc.configuration.value)

    @property
    def device_desc(self):
        return self._device_desc

    # IBootable methods

    def boot(self, session, image, shell_prompt, timeout):
        """
        Boot an image.

        This just starts qemu and waits for the system to settle.

        The session is used to know which qemu to use. Shell prompt
        tells us what to look for, to know that the machine is now
        ready for interactive work.
        """
        assert self._log_file is None
        assert self._proc is None
        args = [
            '-M', self.configuration.qemu_machine,
            # TODO: load this from the backend so that we can boot
            # non-arm images, x86 wants to boot from hdd/fdd/cdrom
            '-drive', 'if=sd,cache=writeback,file={0}'.format(image),
            '-clock', 'unix',
            '-usb',
            '-device', 'usb-kbd',
            '-device', 'usb-mouse',
            '-device', 'usb-net,netdev=mynet',
            # TODO: add more configurability to the network
            # TODO: add support for tap interfaces
            '-netdev', 'user,id=mynet',
            '-nographic']
        software = session.environment.software
        qemu_system = software[self.configuration.qemu_system].load()
        self._log_file = session.scratch.unique(prefix='qemu.', suffix='.log')
        # Start appropriate qemu executable
        self._proc = qemu_system.start(args, logfile=self._log_file)
        # Wait till the machine boots...
        with self.history.debug("Waiting for the shell prompt to appear"):
            # TODO: catch timeout and convert to Timeout
            # TODO: catch EOF and convert to UnrecoverableProblem
            self._proc.expect(shell_prompt, timeout=timeout)

    def stop(self):
        """
        Stop the device
        """
        if self._proc:
            self._proc.close(force=True)
        if self._log_file:
            self._log_file.close()

    # IAgentHost methods

    def start_agent(self):
        """
        Start lava-agent on the device
        """
        assert self._agent is None
        # TODO: capture stuff it can raise and convert to UnrecoverableProblem
        # TODO: catch and convert Timeout
        self._agent = ShellCodeAgent(self._proc)

    @property
    def agent(self):
        """
        Return the device agent instance
        """
        return self._agent

    # TestScenarioInterpreterMixIn methods

    def get_required_software(self):
        return ['linaro-media-create', self.configuration.qemu_system]

    _action_map = {
        'deploy_linaro_image': deploy_linaro_image,
        'boot_linaro_image': boot_linaro_image,
        'lava_test_install': lava_test_install,
        'lava_test_run': lava_test_run,
    }

    def get_action_map(self):
        return self._action_map

    # Other methods

    def __del__(self):
        self.stop()
