# Copyright (C) 2011-2012 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-core
#
# lava-core is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-core.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

"""
lava.core.controllers.devices.rtsm
==================================
"""

import os
import re

from json_document import bridge
from json_document.document import Document
import pexpect

from lava.core.actions.lava import lava_test_install, lava_test_run
from lava.core.actions.linaro import boot_linaro_image
from lava.core.agents import ShellCodeAgent
from lava.core.errors import (SemanticProblem, PreconditionFailed,
                              UnrecoverableProblem, Timeout)
from lava.core.interfaces import IDeviceController, IAgentHost, IBootable
from lava.core.interfaces import ITestScenarioAction
from lava.core.interpreter import TestScenarioInterpreterMixIn
from lava.core.logging import LoggingMixIn
from lava.core.utils import ObjectWithConfiguration, Decoy


__all__ = ["RtsmDeviceController"]


class RTSMImage(object):
    """
    Fast models (currently) have three ways of beeing booted

    1) sd + axf, no semihosting

    Here we just run the model with axf image and the mmc image. Nothing more
    is needed. This mode is used by the integrated switcher.

    2) sd + axf, semihosting, kernel, initrd, device tree

    Here, apart to the files used in 1) we enable semihosting and pass
    kernel/initrd/device tree via semihosting bootloader. This is the reference
    switcher and, in the future, integrated switcher (not really implemented as
    of this time)

    3) sd + axf, semihosting, kernel, initrd, _no_ device tree

    This one is used by KVM, it just carries no device tree.
    """

    # Note: argument names need to be synchronized with `patterns` above and
    # with deployent action parameters below
    def __init__(self, axf_image=None, sd_image=None, kernel_image=None,
                 device_tree=None, initrd=None):
        self.axf_image = axf_image
        self.sd_image = sd_image
        self.kernel_image = kernel_image
        self.device_tree = device_tree
        self.initrd = initrd

    @classmethod
    def guess_from_directory(cls, dirname, patterns):
        """
        Guess the parts of the rtsm image from a directory made by l-m-c

        This function is needed because there is no formal manifest that
        descrbes what each file is for. We're using simple heuristics that look
        at known parts of the file name. If something breaks in the future this
        may need ajusting.
        """
        pieces = {}
        for name in os.listdir(dirname):
            for piece, pattern in patterns.iteritems():
                if re.match(pattern, name):
                    if piece in pieces:
                        raise ValueError(
                            ("Multiple files matched as {0}: {1!r} and"
                             "{2!r}").format(piece, pieces[piece], name))
                    else:
                        pieces[piece] = name
        return cls(**pieces)


class deploy_prebuilt_image(LoggingMixIn, ObjectWithConfiguration,
                            ITestScenarioAction):
    """
    Deploy a prebuilt image composed of:

        1) SD card image
        2) Kernel image
        3) Device tree
        4) Initrd
        5) Special .axf file (bootloader, or bootloader+kernel+switcher magic)
    """

    class Configuration(Document):

        document_schema = {
            'type': 'object',
            'properties': {
                'sd_image': {
                    'type': 'string'
                },
                'kernel_image': {
                    'type': 'string'
                },
                'device_tree': {
                    'type': 'string'
                },
                'initrd': {
                    'type': 'string'
                },
                'axf_image': {
                    'type': 'string'
                }
            },
            'additionalProperties': False,
        }

        @bridge.readonly
        def sd_image(self):
            pass

        @bridge.readonly
        def kernel_image(self):
            pass

        @bridge.readonly
        def device_tree(self):
            pass

        @bridge.readonly
        def initrd(self):
            pass

        @bridge.readonly
        def axf_image(self):
            pass

    def check_precondition(self, state):
        # Check if we are not being run twice
        if 'image' in state:
            raise PreconditionFailed(
                "image is already deployed, deploying twice?")

    def emulate(self, controller, state):
        if not isinstance(controller, RtsmDeviceController):
            raise SemanticProblem("This action is only designed for rtsm")
        state['image'] = RTSMImage(*map(Decoy, (
            "sd.img", "kernel.img", "devicetree.img", "initrd.img",
            "axf.img")))

    def execute(self, controller, session):
        with self.history.info("Downloading prebuilt image pieces"):
            pieces = {}
            for piece in ['sd_image', 'kernel_image', 'device_tree', 'initrd',
                          'axf_image']:
                if piece in self.configuration:
                    pieces[piece] = session.network.download(
                        self.configuration[piece].value, hint=piece)
            session.state['image'] = RTSMImage(**pieces)


class deploy_linaro_image(LoggingMixIn, ObjectWithConfiguration,
                          ITestScenarioAction):
    """
    Deploy an image built by linaro-media-create using rootfs and hwpack
    """

    class Configuration(Document):

        document_schema = {
            'type': 'object',
            'properties': {
                'hwpack': {
                    'type': 'string',
                },
                'rootfs': {
                    'type': 'string',
                },
                'rootfstype': {
                    'type': 'string',
                    'optional': True,
                    "default": "ext4"
                },
                "image_size": {
                    "type": "string",
                    "optional": True,
                    "default": "2G"  # > 2G == broken
                    # TODO: add regexp validation [1G, 2G, 1000-2000M]
                },
                # Map from expected binary type (argument to
                # RTSMImage.__init__()) to the regular expression that
                # describes such files
                "image_piece_patterns": {
                    "type": "object",
                    "default": {},
                    "optional": True,
                    "properties": {
                        "axf_image": {
                            "type": "string",
                            "format": "regex",
                            "default": "^img\\.axf$",
                            "optional": True
                        },
                        "sd_image": {
                            "type": "string",
                            "format": "regex",
                            "default": "^sd\\.img$",
                            "optional": True
                        },
                        "kernel_image": {
                            "type": "string",
                            "format": "regex",
                            "default": "^vmlinuz-.*",
                            "optional": True
                        },
                        "device_tree": {
                            "type": "string",
                            "format": "regex",
                            "default": "\\.dbt$",
                            "optional": True
                        },
                        "initrd": {
                            "type": "string",
                            "format": "regex",
                            "default": "^initrd\\.*",
                            "optional": True
                        }
                    }
                }
            },
            'additionalProperties': False,
        }

        @bridge.readonly
        def rootfs(self):
            """
            URL of the linaro rootfs to deploy
            """

        @bridge.readonly
        def hwpack(self):
            """
            URL of the linaro hwpack to deploy
            """

        @bridge.readonly
        def rootfstype(self):
            """
            Type of the root filesystem to use
            """

        @bridge.readonly
        def image_size(self):
            """
            Size of the combined image

            Typically this is an expression such as 3G for a three gigabyte
            image.
            """

        @bridge.readonly
        def image_piece_patterns(self):
            """
            Dictionary mapping piece name to the regular expression that
            describes files of that type.
            """

    def check_precondition(self, state):
        # Check if we are not being run twice
        if 'image' in state:
            raise PreconditionFailed(
                "image is already deployed, deploying twice?")

    def emulate(self, controller, state):
        if not isinstance(controller, RtsmDeviceController):
            raise SemanticProblem("This action is only designed for rtsm")
        state['image'] = RTSMImage(Decoy("sd.img"), Decoy("img.axf"))

    def execute(self, controller, session):
        # Prepare the image
        with self.history.info("Preparing boot image"):
            # Download both pieces
            rootfs = session.network.download(
                self.configuration.rootfs, hint='rootfs')
            hwpack = session.network.download(
                self.configuration.hwpack, hint="hwpack")
            # And put them together
            image_dir = session.scratch.just_name('fastmodel')
            with session.scratch.open("linaro-media-create.log") as stream:
                with self.history.info(
                    "Combining rootfs and hwpack") as activity:
                    software = session.environment.software
                    lmc = software["linaro-media-create"].load()
                    lmc.generate_image([
                        '--output-directory', image_dir,
                        '--dev', 'fastmodel',
                        '--hwpack', hwpack,
                        '--hwpack-force-yes',
                        '--binary', rootfs,
                        "--rootfs", self.configuration.rootfstype,
                        "--image-size", self.configuration.image_size],
                        logfile=stream)
            # Save image details in session state
            session.state['image'] = RTSMImage.guess_from_directory(
                image_dir, patterns=self.configuration.image_piece_patterns)


class RtsmDeviceController(ObjectWithConfiguration,
                           LoggingMixIn,
                           TestScenarioInterpreterMixIn,
                           IDeviceController, IAgentHost, IBootable):
    """
    Controller for virtual devices implemented with various RTSM (fast model)
    binaries from ARM LTD.
    """

    class Configuration(Document):
        """
        Configuration for the RTSM device controller.

        The actual values are supplied by the device 'configuration' from the
        test environment. They are populated when you instantiate
        RtsmDeviceController.
        """

        document_schema = {
            'type': 'object',
            'properties': {
                'rtsm_binary': {
                    'type': 'string'
                }
            },
            'additionalProperties': False
        }

        @bridge.readonly
        def rtsm_binary(self):
            """
            Name of the RTSM binary executable to use.

            NOTE: this should not be the full executable path. The path is
            defined in environment software.
            """

    def __init__(self, device_desc):
        self._device_desc = device_desc  # DocumentFragment with device data
        self._rtsm = None  # RTSM process (pexpect.spawn)
        self._rtsm_log = None  # log for rtsm
        self._telnet = None  # telnet process (pexpect.spawn)
        self._telnet_log = None  # log for telnet
        self._agent = None   # ShellCodeAgent
        # NOTE: super-call __init__ late as we may crash and call __del__
        ObjectWithConfiguration.__init__(self, device_desc.configuration.value)

    @property
    def device_desc(self):
        return self._device_desc

    # IBootable methods

    def boot(self, session, image, shell_prompt, timeout):
        """
        Boot an image.

        This starts the model binary, waits for it to "go green" and starts
        telnet on an appropriate port.
        """
        # TODO: add support for boot mode (see RTSMImage)
        if not isinstance(image, RTSMImage):
            raise PreconditionFailed(
                "incorrect image, only fast model images are supported")
        assert self._rtsm is None
        assert self._rtsm_log is None
        assert self._telnet is None
        assert self._telnet_log is None
        args = [
            "-a", "coretile.cluster0.*={0}".format(image.axf_image),
            "-C", "motherboard.smsc_91c111.enabled=1",
            "-C", "motherboard.hostbridge.userNetworking=1",
            "-C", "motherboard.mmc.p_mmc_file={0}".format(image.sd_image),
            # TODO: make this configurable, depending on boot mode
            "-C", "coretile.cluster0.cpu0.semihosting-enable=1",
            "-C", "motherboard.vis.disable_visualisation=1",
            "-C", "motherboard.terminal_0.start_telnet=0",
            "-C", "motherboard.terminal_1.start_telnet=0",
            "-C", "motherboard.terminal_2.start_telnet=0",
            "-C", "motherboard.terminal_3.start_telnet=0",
            # TODO: make this configurable
            "-C", "coretile.cache_state_modelled=1",
            # TODO: vary this on boot mode and parts of the image we really
            # have. In particular kvm does not have device tree
            "-C", ("coretile.cluster0.cpu0.semihosting-cmd_line=\""
                   "--kernel {0} --initrd {1} --fdt {2}"
                   " -- "
                   # everything after this goes to the kernel (not to
                   # semihosting bootloader)
                   "mem=2046M "  # TODO: make this configurable
                   "console=ttyAMA0,115200 "
                   "migration_cost=500 "
                   "cachepolicy=writealloc "
                   "root=/dev/mmcblk0p2").format(
                       image.kernel_image,
                       image.initrd,
                       image.device_tree)]
        software = session.environment.software
        # Start appropriate rtsm executable
        with self.history.info("Starting fast model {rtsm_binary}",
                               rtsm_binary=self.configuration.rtsm_binary) as activity:
            rtsm_binary = software[self.configuration.rtsm_binary].load()
            self._rtsm_log = session.scratch.unique(prefix='rtsm.', suffix='.log')
            self._rtsm = rtsm_binary.start(args, logfile=self._rtsm_log)
        # Wait till simulation starts
        with self.history.debug("Waiting for the simulator to start"):
            match = self._rtsm.expect(["ERROR: License check failed!",
                                       "Simulation is started"])
            if match == 0:
                raise UnrecoverableProblem("fast model license check failed")
        # Start telnet and connect to rtsm
        # XXX: when we get to run many vms on a single host we need to start
        # randomizing ports, this is bad as we always get to have a race
        # condition we may either need to do some kludge and work around it or
        # try to transition to non-socket communication wit the model
        # TODO: maybe attach to the other three serial lines?
        with self.history.info(
            "Using telnet to connect to running machine") as activity:
            telnet_binary = software['telnet'].load()
            self._telnet_log = session.scratch.unique(
                prefix='telnet.0.', suffix='.log')
            self._telnet_log.close()  # just the unique file name please
            self._telnet = telnet_binary.connect_to('localhost', 5000, self._telnet_log.name)
            activity.attach(self._telnet_log.name)
        # Wait till the machine boots...
        with self.history.debug("Waiting for the shell prompt to appear"
                                " (timeout: {timeout}", timeout=timeout):
            match = self._telnet.expect(
                [pexpect.EOF, pexpect.TIMEOUT, shell_prompt],
                timeout=timeout)
            if match == 0:
                raise UnrecoverableError("telnet died before we finished booting")
            elif match == 1:
                raise Timeout("Failed to boot in {0} seconds".format(timeout))

    def stop(self):
        """
        Stop the device
        """
        if self._rtsm:
            import pexpect
            with self.history.debug("Shutting down model process"):
                # Send ctrl-c and wait for the thing to exit
                self._rtsm.sendintr()
                self._rtsm.expect(pexpect.EOF)
                self._rtsm.close()
                self._rtsm = None
        if self._rtsm_log:
            self._rtsm_log.close()
            self._rtsm = None
        if self._telnet:
            self.history.debug("Shutting down telnet process")
            self._telnet.close(force=True)
            self._telnet = None

    # IAgentHost methods

    def start_agent(self):
        """
        Start lava-agent on the device
        """
        assert self._agent is None
        assert self._telnet is not None
        # TODO: capture stuff it can raise and convert to UnrecoverableProblem
        # TODO: catch and convert Timeout
        self._agent = ShellCodeAgent(self._telnet)

    @property
    def agent(self):
        """
        Return the device agent instance
        """
        return self._agent

    # TestScenarioInterpreterMixIn methods

    def get_required_software(self):
        return [
            'linaro-media-create',
            'linaro-hwpack-create',
            'telnet',
            self.configuration.rtsm_binary]

    _action_map = {
        'deploy_prebuilt_image': deploy_prebuilt_image,
        'deploy_linaro_image': deploy_linaro_image,
        'boot_linaro_image': boot_linaro_image,
        'lava_test_install': lava_test_install,
        'lava_test_run': lava_test_run,
    }

    def get_action_map(self):
        return self._action_map

    # Other methods

    def __del__(self):
        self.stop()
